import numpy as np
import pandas as pd
from skfolio import Portfolio as BasePortfolio

from .utils import fix_quantization_error


class Portfolio(BasePortfolio):
    @property
    def all_weights_per_observation(self) -> pd.DataFrame:
        """DataFrame of the Portfolio weights containing even near zero weight per observation."""
        weights = self.weights
        assets = self.assets
        df = pd.DataFrame(
            np.ones((len(self.observations), len(assets))) * weights,
            index=self.observations,
            columns=assets,
        )
        return df

    def get_contributions(self) -> tuple[pd.Series, float]:
        returns = self.X.iloc[-1, :].T
        weights = self.all_weights_per_observation.iloc[-1, :].T
        portfolio_returns = (weights * (returns + 1.0)).sum()
        return returns, portfolio_returns

    def get_next_weights(self, round_precision: int = 8) -> dict[int, float]:
        """
        Given the next returns, compute the drifted weights of this portfolio
            round_precision: Round the weight to the given round number and ensure the total weight reflects this. Default to 8 decimals
        Returns:
            A dictionary of weights (instrument ids as keys and weights as values)
        """
        returns, portfolio_returns = self.get_contributions()
        weights = self.all_weights_per_observation.iloc[-1, :].T
        next_weights = weights * (returns + 1.0) / portfolio_returns
        next_weights = next_weights.dropna()
        next_weights = next_weights / next_weights.sum()
        if round_precision and not next_weights.empty:
            next_weights = fix_quantization_error(next_weights, round_precision)
        return {i: round(w, round_precision) for i, w in next_weights.items()}  # handle float precision manually

    def get_estimate_net_value(self, previous_net_asset_value: float) -> float:
        expected_returns = self.weights @ self.X.iloc[-1, :].T
        return previous_net_asset_value * (1.0 + expected_returns)



from django.shortcuts import get_object_or_404
from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbportfolio.models import OrderProposal


class OrderProposalEndpointConfig(EndpointViewConfig):
    def get_delete_endpoint(self, **kwargs):
        if pk := self.view.kwargs.get("pk", None):
            order_proposal = get_object_or_404(OrderProposal, pk=pk)
            if order_proposal.status in [OrderProposal.Status.DRAFT, OrderProposal.Status.DENIED]:
                return super().get_endpoint()
        return None


class OrderProposalPortfolioEndpointConfig(OrderProposalEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-orderproposal-list", args=[self.view.kwargs["portfolio_id"]], request=self.request
        )



from wbcore.menus import ItemPermission, MenuItem

TRADE_MENUITEM = MenuItem(
    label="Trades",
    endpoint="wbportfolio:trade-list",
    permission=ItemPermission(permissions=["wbportfolio.view_trade"], method=lambda request: request.user.is_internal),
)
SUBSCRIPTION_REDEMPTION_MENUITEM = MenuItem(
    label="Subscription/Redemption",
    endpoint="wbportfolio:subscriptionredemption-list",
    permission=ItemPermission(permissions=["wbportfolio.view_trade"], method=lambda request: request.user.is_internal),
)



from django.conf import settings
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.models import Company, Entry, Person
from wbcore.contrib.directory.serializers.companies import (
    CompanyModelListSerializer as BaseCompanyModelListSerializer,
)
from wbcore.contrib.directory.serializers.companies import (
    CompanyModelSerializer as BaseCompanyModelSerializer,
)
from wbcore.contrib.directory.serializers.entries import (
    CompanyRepresentationSerializer as BaseCompanyRepresentationSerializer,
)
from wbcore.contrib.directory.serializers.persons import (
    PersonModelListSerializer as BasePersonModelListSerializer,
)
from wbcore.contrib.directory.serializers.persons import (
    PersonModelSerializer as BasePersonModelSerializer,
)
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer
from wbcore.serializers.fields.types import DisplayMode

from .constants import DEFAULT_PORTFOLIO_DATA_UPDATE_TASK_INTERVAL_MINUTES
from .models import (
    AssetAllocation,
    AssetAllocationType,
    CompanyPortfolioData,
    GeographicFocus,
)

HELP_TEXT = f" (Updated every {DEFAULT_PORTFOLIO_DATA_UPDATE_TASK_INTERVAL_MINUTES} minutes)"


def _get_assets_under_management_kwargs(field_name):
    field = getattr(CompanyPortfolioData, field_name).field
    return {
        "max_digits": field.max_digits,
        "decimal_places": field.decimal_places,
        "label": field.verbose_name,
        "help_text": field.help_text + HELP_TEXT,
        "display_mode": DisplayMode.SHORTENED,
        "required": False,
    }


def _get_investment_discretion_kwargs():
    field = CompanyPortfolioData.investment_discretion.field
    return {
        "choices": field.choices,
        "default": field.default,
        "help_text": field.help_text + HELP_TEXT,
        "label": field.verbose_name,
    }


def _get_potential_kwargs():
    return {
        "decimal_places": CompanyPortfolioData.potential.field.decimal_places,
        "max_digits": CompanyPortfolioData.potential.field.max_digits,
        "help_text": CompanyPortfolioData.potential.field.help_text + HELP_TEXT,
        "display_mode": DisplayMode.SHORTENED,
        "read_only": True,
    }


def get_portfolio_data(validated_data, method="get"):
    method = getattr(validated_data, method)

    assets_under_management = method("asset_under_management", None)
    assets_under_management_currency_id = method("assets_under_management_currency", None)
    investment_discretion = method("investment_discretion", None)

    if isinstance(assets_under_management_currency_id, Currency):
        assets_under_management_currency_id = assets_under_management_currency_id.id

    return validated_data, {
        "assets_under_management": assets_under_management,
        "assets_under_management_currency_id": assets_under_management_currency_id,
        "investment_discretion": investment_discretion,
    }


def update_portfolio_data(company_portfolio_data, portfolio_data):
    if assets_under_management := portfolio_data["assets_under_management"]:
        company_portfolio_data.assets_under_management = assets_under_management

    if assets_under_management_currency_id := portfolio_data["assets_under_management_currency_id"]:
        company_portfolio_data.assets_under_management_currency_id = assets_under_management_currency_id

    if investment_discretion := portfolio_data["investment_discretion"]:
        company_portfolio_data.investment_discretion = investment_discretion
    company_portfolio_data.update()
    company_portfolio_data.save()


class CompanyPortfolioDataMixin(serializers.ModelSerializer):
    asset_under_management = serializers.DecimalField(
        **_get_assets_under_management_kwargs(field_name="assets_under_management"),
    )
    invested_assets_under_management_usd = serializers.DecimalField(
        read_only=True,
        **_get_assets_under_management_kwargs(field_name="invested_assets_under_management_usd"),
    )

    assets_under_management_currency_repr = serializers.CharField(read_only=True, required=False)

    potential = serializers.DecimalField(**_get_potential_kwargs())
    investment_discretion = serializers.ChoiceField(**_get_investment_discretion_kwargs())

    # Not sure why read_only_fields does not work...
    tier = serializers.ChoiceField(
        read_only=True, choices=Company.Tiering.choices, help_text=settings.DEFAULT_TIERING_HELP_TEXT
    )

    class Meta:
        model = Entry
        read_only_fields = ("tier",)
        decorators = {
            "asset_under_management": serializers.decorator(
                position="left", value="{{assets_under_management_currency_repr}}"
            ),
            "potential": serializers.decorator(
                decorator_type="text", position="left", value="{{assets_under_management_currency_repr}}"
            ),
            "invested_assets_under_management_usd": serializers.decorator(
                decorator_type="text", position="left", value="$"
            ),
        }

        fields = (
            "asset_under_management",  # TODO: add an s after asset - After removing this field from the base model
            "assets_under_management_currency_repr",
            "invested_assets_under_management_usd",
            "investment_discretion",
            "potential",
            "tier",
        )


class CompanyModelSerializer(CompanyPortfolioDataMixin, BaseCompanyModelSerializer):
    SERIALIZER_CLASS_FOR_REMOTE_ADDITIONAL_RESOURCES = BasePersonModelSerializer
    assets_under_management_currency = serializers.PrimaryKeyRelatedField(
        required=False,
        queryset=Currency.objects.all(),
        default=None,
        label=CompanyPortfolioData.assets_under_management_currency.field.verbose_name,
    )
    _assets_under_management_currency = CurrencyRepresentationSerializer(source="assets_under_management_currency")
    potential_currency = serializers.PrimaryKeyRelatedField(
        required=False,
        queryset=Currency.objects.all(),
        label=CompanyPortfolioData.potential_currency.field.verbose_name,
    )
    _potential_currency = CurrencyRepresentationSerializer(source="potential_currency")

    def update(self, instance, validated_data):
        validated_data, portfolio_data = get_portfolio_data(validated_data)
        updated_instance = super().update(instance, validated_data)
        update_portfolio_data(updated_instance.portfolio_data, portfolio_data)

        return updated_instance

    def create(self, validated_data):
        validated_data, portfolio_data = get_portfolio_data(validated_data, method="pop")
        created_instance = super().create(validated_data)
        update_portfolio_data(created_instance.portfolio_data, portfolio_data)

        return created_instance

    @serializers.register_resource()
    def extra_additional_resources(self, instance, request, user):
        return {
            "asset_allocation_table": reverse(
                "company_portfolio:companyassetallocation-list", kwargs={"company_id": instance.id}, request=request
            ),
            "geographic_focus_table": reverse(
                "company_portfolio:companygeographicfocus-list", kwargs={"company_id": instance.id}, request=request
            ),
        }

    class Meta(CompanyPortfolioDataMixin.Meta, BaseCompanyModelSerializer.Meta):
        model = Company
        read_only_fields = (
            *BaseCompanyModelSerializer.Meta.read_only_fields,
            *CompanyPortfolioDataMixin.Meta.read_only_fields,
        )
        fields = (
            *BaseCompanyModelSerializer.Meta.fields,
            *CompanyPortfolioDataMixin.Meta.fields,
            "assets_under_management_currency",
            "_assets_under_management_currency",
            "potential_currency",
            "_potential_currency",
        )


class CompanyModelListSerializer(CompanyPortfolioDataMixin, BaseCompanyModelListSerializer):
    SERIALIZER_CLASS_FOR_REMOTE_ADDITIONAL_RESOURCES = BaseCompanyModelListSerializer

    class Meta(CompanyPortfolioDataMixin.Meta, BaseCompanyModelListSerializer.Meta):
        model = Company
        fields = (*BaseCompanyModelListSerializer.Meta.fields, *CompanyPortfolioDataMixin.Meta.fields)


class PersonModelSerializer(CompanyPortfolioDataMixin, BasePersonModelSerializer):
    SERIALIZER_CLASS_FOR_REMOTE_ADDITIONAL_RESOURCES = BasePersonModelSerializer

    assets_under_management_currency = serializers.PrimaryKeyRelatedField(
        required=False,
        queryset=Currency.objects.all(),
        default=None,
        label=CompanyPortfolioData.assets_under_management_currency.field.verbose_name,
    )
    _assets_under_management_currency = CurrencyRepresentationSerializer(source="assets_under_management_currency")
    potential_currency = serializers.PrimaryKeyRelatedField(
        required=False,
        queryset=Currency.objects.all(),
        label=CompanyPortfolioData.potential_currency.field.verbose_name,
    )
    _potential_currency = CurrencyRepresentationSerializer(source="potential_currency")

    asset_under_management = serializers.DecimalField(
        **_get_assets_under_management_kwargs(field_name="assets_under_management"),
        read_only=True,
    )
    investment_discretion = serializers.ChoiceField(**_get_investment_discretion_kwargs(), read_only=True)

    class Meta(CompanyPortfolioDataMixin.Meta, BasePersonModelSerializer.Meta):
        model = Person
        read_only_fields = (
            *BasePersonModelSerializer.Meta.read_only_fields,
            *CompanyPortfolioDataMixin.Meta.fields,
            *CompanyPortfolioDataMixin.Meta.read_only_fields,
        )
        fields = (
            *BasePersonModelSerializer.Meta.fields,
            *CompanyPortfolioDataMixin.Meta.fields,
            "assets_under_management_currency",
            "_assets_under_management_currency",
            "potential_currency",
            "_potential_currency",
        )


class PersonModelListSerializer(CompanyPortfolioDataMixin, BasePersonModelListSerializer):
    investment_discretion = serializers.ChoiceField(**_get_investment_discretion_kwargs(), read_only=True)
    SERIALIZER_CLASS_FOR_REMOTE_ADDITIONAL_RESOURCES = BasePersonModelListSerializer

    class Meta(CompanyPortfolioDataMixin.Meta, BasePersonModelListSerializer.Meta):
        model = Person
        fields = (*BasePersonModelListSerializer.Meta.fields, *CompanyPortfolioDataMixin.Meta.fields)
        read_only_fields = fields


class AssetAllocationTypeRepresentationSerializer(serializers.RepresentationSerializer):
    class Meta:
        model = AssetAllocationType
        fields = ("id", "name")


class AssetAllocationTypeModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = AssetAllocationType
        fields = ("id", "name", "_additional_resources")


class AssetAllocationModelSerializer(serializers.ModelSerializer):
    _asset_type = AssetAllocationTypeRepresentationSerializer(source="asset_type")
    _company = BaseCompanyRepresentationSerializer(source="company")
    company = serializers.PrimaryKeyRelatedField(
        default=serializers.DefaultFromGET("company"), queryset=Company.objects.all(), label="Company"
    )
    comment = serializers.TextAreaField(default="")

    class Meta:
        percent_fields = ("percent", "max_investment")
        model = AssetAllocation
        fields = (
            "id",
            "company",
            "_company",
            "asset_type",
            "_asset_type",
            "percent",
            "max_investment",
            "comment",
            "_additional_resources",
        )


class GeographicFocusModelSerializer(serializers.ModelSerializer):
    _country = GeographyRepresentationSerializer(source="country")
    company = serializers.PrimaryKeyRelatedField(
        default=serializers.DefaultFromGET("company"), queryset=Company.objects.all(), label="Company"
    )
    _company = BaseCompanyRepresentationSerializer(source="company")
    comment = serializers.TextAreaField(default="")

    class Meta:
        percent_fields = ("percent",)
        model = GeographicFocus
        fields = (
            "id",
            "company",
            "_company",
            "country",
            "_country",
            "percent",
            "comment",
            "_additional_resources",
        )



import datetime
import logging
import re

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import convert_string_to_number

logger = logging.getLogger("importers.parsers.jp_morgan.strategy")


def file_name_parse(file_name):
    dates = re.findall("([0-9]{8})", file_name)
    if dates:
        return {"valuation_date": datetime.datetime.strptime(dates[0], "%Y%m%d").date()}
    return {}


def parse(import_source):
    # Load file into a CSV DictReader

    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=",")

    # Parse the Parts of the filename into the different parts
    parts = file_name_parse(import_source.file.name)

    # Get the valuation date from the parts list
    report_date = parts.get("valuation_date")

    # Iterate through the CSV File and parse the data into a list
    data = list()
    if "Date" in df.columns:
        df["Date"] = pd.to_datetime(df["Date"])
    df.replace([np.inf, -np.inf, np.nan], None, inplace=True)
    for strategy_data in df.to_dict("records"):
        valuation_date = strategy_data.get("Date", report_date)
        if valuation_date:
            bbg_tickers = strategy_data["BBG Ticker"].split(" ")
            exchange = None
            if len(bbg_tickers) == 2:
                ticker = bbg_tickers[0]
                instrument_type = bbg_tickers[1]
            elif len(bbg_tickers) == 3:
                ticker = bbg_tickers[0]
                exchange = bbg_tickers[1]
                instrument_type = bbg_tickers[2]

            strategy = strategy_data["Strategy Ticker"].replace("Index", "").strip()
            strategy_currency_key = strategy_data["Strategy CCY"]

            position_currency_key = strategy_data["Position CCY"]

            isin = strategy_data["Position ISIN"]
            name = strategy_data["Position Description"]
            initial_price = convert_string_to_number(strategy_data["Prices"])
            initial_currency_fx_rate = convert_string_to_number(strategy_data["Fx Rates"])
            if exchange:
                exchange = {"bbg_exchange_codes": exchange}
            try:
                weighting = convert_string_to_number(strategy_data["Weight In Percent"].replace("%", "")) / 100
            except Exception:
                weighting = 0.0
            underlying_quote = {
                "ticker": ticker,
                "exchange": exchange,
                "isin": isin,
                "name": name,
                "currency__key": position_currency_key,
                "instrument_type": instrument_type.lower(),
            }
            if isin:
                underlying_quote["isin"] = isin
            data.append(
                {
                    "underlying_quote": underlying_quote,
                    "portfolio": {
                        "instrument_type": "index",
                        "ticker": strategy,
                        "currency__key": strategy_currency_key,
                    },
                    "exchange": exchange,
                    "is_estimated": False,
                    "currency__key": position_currency_key,
                    "initial_currency_fx_rate": initial_currency_fx_rate,
                    "weighting": weighting,
                    "initial_price": initial_price,
                    "date": valuation_date.strftime("%Y-%m-%d"),
                }
            )
    return {"data": data}



from collections import defaultdict
from datetime import date, timedelta
from decimal import Decimal

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from django.contrib.messages import info
from django.db.models import (
    ExpressionWrapper,
    F,
    FloatField,
    OuterRef,
    Q,
    Subquery,
    Sum,
)
from django.db.models.functions import Coalesce
from django.dispatch import receiver
from django.utils.functional import cached_property
from wbcore import viewsets
from wbcore.cache.decorators import cache_table
from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin
from wbcore.signals.instance_buttons import add_instance_button
from wbcore.utils.date import (
    get_date_interval_from_request,
    get_next_day_timedelta,
    get_start_and_end_date_from_date,
)
from wbcore.utils.figures import (
    get_default_timeserie_figure,
    get_hovertemplate_timeserie,
)
from wbcore.utils.strings import format_number
from wbfdm.contrib.metric.backends.performances import PERFORMANCE_METRIC
from wbfdm.contrib.metric.viewsets.mixins import InstrumentMetricMixin
from wbfdm.models import Instrument, InstrumentPrice

from wbportfolio import serializers
from wbportfolio.filters import (
    BaseProductFilterSet,
    ProductCustomerFilter,
    ProductFeeFilter,
    ProductFilter,
)
from wbportfolio.models import Fees, Product, Trade
from wbportfolio.viewsets.configs.buttons.mixins import InstrumentButtonMixin

from .configs import (
    AUMProductEndpointConfig,
    AUMProductTitleConfig,
    InstrumentPriceAUMDataEndpointConfig,
    InstrumentPriceAUMDataTitleConfig,
    NominalProductEndpointConfig,
    NominalProductTitleConfig,
    ProductButtonConfig,
    ProductCustomerButtonConfig,
    ProductCustomerDisplayConfig,
    ProductCustomerEndpointConfig,
    ProductDisplayConfig,
    ProductPerformanceFeesDisplayConfig,
    ProductPerformanceFeesEndpointConfig,
    ProductPerformanceFeesTitleConfig,
)
from .mixins import UserPortfolioRequestPermissionMixin


class ProductRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:product"
    serializer_class = serializers.ProductRepresentationSerializer
    filterset_class = BaseProductFilterSet

    queryset = Product.objects.all()
    ordering_fields = ordering = ("computed_str",)
    search_fields = ("computed_str",)

    def get_queryset(self):
        return Product.get_products(self.request.user.profile).annotate(bank_name=F("bank__name"))


class ProductModelViewSet(InstrumentMetricMixin, viewsets.ModelViewSet):
    filterset_class = ProductFilter
    queryset = Product.objects.all()
    METRIC_KEYS = (PERFORMANCE_METRIC,)
    METRIC_WITH_PREFIXED_KEYS = True

    @property
    def metric_basket_class(self):
        return Instrument

    ordering_fields = (
        "is_invested",
        "name_repr",
        "parent__name__nulls_last",
        "bank__name",
        "current_issuer_fees_percent",
        "currency__key",
        "isin",
        "ticker",
        "last_valuation_date",
        "net_value",
        "assets_under_management",
        "assets_under_management_usd",
        "nnm_weekly",
        "nnm_monthly",
        "nnm_year_to_date",
        "nnm_yearly",
        "tags__title__nulls_last",
    )
    search_fields = (
        "name",
        "isin",
        "ticker",
        "white_label_customers__computed_str",
        "parent__name",
        "computed_str",
    )
    ordering = ["name_repr"]

    display_config_class = ProductDisplayConfig
    button_config_class = ProductButtonConfig

    def get_aggregates(self, queryset, paginated_queryset):
        aggregates = {}
        if self.request.user.is_internal:
            df = pd.DataFrame(
                queryset.values(
                    "currency",
                    "nnm_weekly",
                    "nnm_monthly",
                    "nnm_year_to_date",
                    "nnm_yearly",
                    "assets_under_management",
                    "assets_under_management_usd",
                    "is_invested",
                )
            )  # we use pandas to avoid unnecessary db calls
            product_aggregates = defaultdict(dict)
            if not df.empty:
                currency_symbol_map = dict(Currency.objects.filter(id__in=df.currency).values_list("id", "symbol"))
                for currency_id in df.currency.unique():
                    currency_symbol = currency_symbol_map[currency_id]

                    for field in ["nnm_weekly", "nnm_monthly", "nnm_year_to_date", "nnm_yearly"]:
                        product_aggregates[field][currency_symbol] = format_number(
                            df.loc[df["currency"] == currency_id, field].sum()
                        )

                    dff = df.loc[(df["currency"] == currency_id) & (df["is_invested"])]
                    product_aggregates["assets_under_management"][currency_symbol] = format_number(
                        dff["assets_under_management"].sum()
                    )
                product_aggregates["assets_under_management_usd"]["Σ"] = format_number(
                    df[df["is_invested"]]["assets_under_management_usd"].sum()
                )

                # Double Accounting with only is_invested = False
                product_aggregates["assets_under_management_usd"]["Σ (DA)"] = format_number(
                    df[~df["is_invested"]]["assets_under_management_usd"].sum()
                )

            aggregates.update(product_aggregates)

        return aggregates

    def get_serializer_class(self):
        if getattr(self, "action", None) == "list":
            return serializers.ProductListModelSerializer
        return serializers.ProductModelSerializer

    def get_queryset(self):
        today = date.today()
        base_qs = Product.annotate_last_aum(
            Product.get_products(self.request.user.profile, base_qs=super().get_queryset())
        )
        return (
            base_qs.annotate(
                nnm_weekly=Coalesce(
                    Subquery(
                        Trade.valid_external_customer_trade_objects.filter(
                            underlying_instrument=OuterRef("pk"), transaction_date__gte=today - timedelta(days=7)
                        )
                        .values("underlying_instrument")
                        .annotate(sum_aum=Sum("total_value"))
                        .values("sum_aum")[:1]
                    ),
                    Decimal(0),
                ),
                nnm_monthly=Coalesce(
                    Subquery(
                        Trade.valid_external_customer_trade_objects.filter(
                            underlying_instrument=OuterRef("pk"), transaction_date__gte=today - timedelta(days=30)
                        )
                        .values("underlying_instrument")
                        .annotate(sum_aum=Sum("total_value"))
                        .values("sum_aum")[:1]
                    ),
                    Decimal(0),
                ),
                nnm_year_to_date=Coalesce(
                    Subquery(
                        Trade.valid_external_customer_trade_objects.filter(
                            underlying_instrument=OuterRef("pk"), transaction_date__year=today.year
                        )
                        .values("underlying_instrument")
                        .annotate(sum_aum=Sum("total_value"))
                        .values("sum_aum")[:1]
                    ),
                    Decimal(0),
                ),
                nnm_yearly=Coalesce(
                    Subquery(
                        Trade.valid_external_customer_trade_objects.filter(
                            underlying_instrument=OuterRef("pk"), transaction_date__gte=today - timedelta(days=365)
                        )
                        .values("underlying_instrument")
                        .annotate(sum_aum=Sum("total_value"))
                        .values("sum_aum")[:1]
                    ),
                    Decimal(0),
                ),
            )
            .select_related("currency", "bank", "parent")
            .prefetch_related("white_label_customers", "classifications")
        )


class ProductCustomerModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbportfolio:productcustomer"

    serializer_class = serializers.ProductCustomerModelSerializer
    filterset_class = ProductCustomerFilter
    queryset = Product.active_objects.all()

    search_fields = ("name", "isin", "ticker")
    ordering_fields = ("name", "isin", "ticker")
    ordering = ["name"]

    display_config_class = ProductCustomerDisplayConfig
    button_config_class = ProductCustomerButtonConfig
    endpoint_config_class = ProductCustomerEndpointConfig

    def get_queryset(self):
        qs = (
            Product.get_products(self.request.user.profile)
            .filter(id__in=Product.active_objects.values("id"))
            .annotate(
                net_value=InstrumentPrice.subquery_closest_value(
                    "net_value",
                    val_date=None,
                    date_name="last_valuation_date",
                    instrument_pk_name="pk",
                    date_lookup="exact",
                ),
                bank_repr=F("bank__name"),
                currency_symbol=F("currency__symbol"),
            )
        )
        return qs.filter(net_value__isnull=False)


class ProductPerformanceFeesModelViewSet(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, viewsets.ModelViewSet
):
    search_fields = ("name", "isin", "ticker", "computed_str")
    ordering_fields = (
        "computed_str",
        "isin",
        "ticker",
        "sum_management_fees",
        "sum_management_fees_usd",
        "sum_performance_fees_net",
        "sum_total",
        "sum_performance_fees_net_usd",
        "sum_total_usd",
    )
    ordering = ["computed_str"]
    queryset = Product.active_objects.all()

    serializer_class = serializers.ProductFeesModelSerializer
    filterset_class = ProductFeeFilter

    display_config_class = ProductPerformanceFeesDisplayConfig
    title_config_class = ProductPerformanceFeesTitleConfig
    endpoint_config_class = ProductPerformanceFeesEndpointConfig

    def get_aggregates(self, queryset, paginated_queryset):
        if not queryset.exists():
            return dict()
        return {
            # NOTE: This does not work at the moment. Somehow grouped by currency is a problem...Weird
            "sum_management_fees_usd": {
                "Σ": format_number(queryset.aggregate(s=Sum(F("sum_management_fees_usd")))["s"])
            },
            "sum_performance_fees_net_usd": {
                "Σ": format_number(queryset.aggregate(s=Sum(F("sum_performance_fees_net_usd")))["s"])
            },
            "sum_total_usd": {"Σ": format_number(queryset.aggregate(s=Sum(F("sum_total_usd")))["s"])},
        }

    @cached_property
    def latest_rate_date(self) -> date:
        try:
            return CurrencyFXRates.objects.latest("date").date
        except CurrencyFXRates.DoesNotExist:
            return date.today()

    def add_messages(self, request, queryset=None, paginated_queryset=None, instance=None, initial=False):
        return info(request, f"The FX Rates from the {self.latest_rate_date:%d.%m.%Y} were used.")

    def get_queryset(self):
        if (latest_rate_date := self.latest_rate_date) and self.is_manager:
            # Check for Filter because they are fake
            date_gte, date_lte = get_date_interval_from_request(self.request, exclude_weekend=True)

            if date_lte is None and date_gte is None:
                date_gte, date_lte = get_start_and_end_date_from_date(date.today())

            fees = Fees.valid_objects.filter(product=OuterRef("pk"))
            if date_gte:
                fees = fees.filter(fee_date__gte=date_gte)
            if date_lte:
                fees = fees.filter(fee_date__lte=date_lte)

            management_fees = Coalesce(
                Subquery(
                    fees.filter(transaction_subtype=Fees.Type.MANAGEMENT)
                    .values("product")
                    .annotate(sum_management_fees=Sum("total_value"))
                    .values("sum_management_fees")[:1],
                    output_field=FloatField(),
                ),
                0.0,
            )

            performance_fees_net = Coalesce(
                Subquery(
                    fees.filter(
                        Q(transaction_subtype=Fees.Type.PERFORMANCE)
                        | Q(transaction_subtype=Fees.Type.PERFORMANCE_CRYSTALIZED)
                    )
                    .values("product")
                    .annotate(sum_performance_fees_net=Sum("total_value"))
                    .values("sum_performance_fees_net")[:1],
                    output_field=FloatField(),
                ),
                0.0,
            )

            qs = (
                Product.get_products(self.request.user.profile)
                .annotate(
                    sum_management_fees=management_fees,
                    sum_performance_fees_net=performance_fees_net,
                    fx_rate=CurrencyFXRates.get_fx_rates_subquery(
                        latest_rate_date, currency="currency", lookup_expr="exact"
                    ),
                    sum_management_fees_usd=ExpressionWrapper(
                        F("sum_management_fees") * F("fx_rate"), output_field=FloatField()
                    ),
                    sum_performance_fees_net_usd=ExpressionWrapper(
                        F("sum_performance_fees_net") * F("fx_rate"),
                        output_field=FloatField(),
                    ),
                    sum_total=F("sum_management_fees") + F("sum_performance_fees_net"),
                    sum_total_usd=F("sum_management_fees_usd") + F("sum_performance_fees_net_usd"),
                )
                .select_related("currency")
                .prefetch_related("fees", "prices")
            )
            return qs

        return Product.objects.none()


# Product based InstrumentPrice Chartview sets


class NominalProductChartView(viewsets.ChartViewSet):
    IDENTIFIER = "wbportfolio:price"
    filterset_fields = {
        "date": ["gte", "exact", "lte"],
    }
    queryset = InstrumentPrice.objects.all()

    title_config_class = NominalProductTitleConfig
    endpoint_config_class = NominalProductEndpointConfig

    def get_plotly(self, queryset):
        fig = get_default_timeserie_figure()
        product = Product.objects.get(id=self.kwargs["product_id"])
        df = pd.DataFrame(queryset.values("date", "outstanding_shares", "calculated"))
        df = df.sort_values(by="calculated").groupby("date").first().dropna()
        df["nominal_value"] = df["outstanding_shares"] * product.share_price
        fig.add_trace(
            go.Scatter(
                x=df.index,
                y=df.nominal_value,
                mode="lines",
                fill="tozeroy",
                name=f"Nominal Value ({product.currency.key})",
                hovertemplate=get_hovertemplate_timeserie(currency=""),
            )
        )
        return fig

    def get_queryset(self):
        return InstrumentPrice.objects.filter(instrument=self.kwargs["product_id"], outstanding_shares__isnull=False)


class AUMProductChartView(viewsets.ChartViewSet):
    IDENTIFIER = "wbportfolio:price"
    filterset_fields = {
        "date": ["gte", "exact", "lte"],
    }
    queryset = InstrumentPrice.objects.all()

    title_config_class = AUMProductTitleConfig
    endpoint_config_class = AUMProductEndpointConfig

    def get_plotly(self, queryset):
        fig = get_default_timeserie_figure()
        product = Product.objects.get(id=self.kwargs["product_id"])
        df = pd.DataFrame(queryset.values("date", "outstanding_shares", "net_value", "calculated"))
        df = df.sort_values(by="calculated").groupby("date").first().dropna()
        df["assets_under_management"] = df["outstanding_shares"] * df["net_value"]

        fig.add_trace(
            go.Scatter(
                x=df.index,
                y=df.assets_under_management,
                mode="lines",
                fill="tozeroy",
                name=f"AUM ({product.currency.key})",
                hovertemplate=get_hovertemplate_timeserie(currency=""),
            )
        )

        return fig

    def get_queryset(self):
        return super().get_queryset().filter(instrument=self.kwargs["product_id"], outstanding_shares__isnull=False)


@cache_table(timeout=get_next_day_timedelta(), periodic_caching=True)
class InstrumentPriceAUMDataChartView(viewsets.ChartViewSet):
    IDENTIFIER = "wbportfolio:aum-price"
    LIST_TITLE = "Assets under Management"

    queryset = InstrumentPrice.objects.annotate_base_data()

    title_config_class = InstrumentPriceAUMDataTitleConfig
    endpoint_config_class = InstrumentPriceAUMDataEndpointConfig

    def get_plotly(self, queryset):
        fig = get_default_timeserie_figure(add_rangeslider=False)

        df_net_value = pd.DataFrame(
            queryset.filter(calculated=False, net_value_usd__isnull=False)
            .exclude(net_value_usd=0)
            .values("instrument", "date", "net_value_usd")
        )
        df_outstanding_shares = pd.DataFrame(
            queryset.filter(outstanding_shares__isnull=False)
            .exclude(outstanding_shares=0)
            .values("instrument", "date", "outstanding_shares", "calculated")
        )
        if not df_outstanding_shares.empty and not df_net_value.empty:
            # get the none calculated outstanding share if possible, otherwise get the calculated one
            df_outstanding_shares = (
                (
                    df_outstanding_shares.sort_values(by=["instrument", "date", "calculated"])
                    .groupby(["instrument", "date"])
                    .first()
                )
                .groupby(level=0)
                .ffill()
            )
            df_net_value = df_net_value.set_index(["instrument", "date"])
            df = pd.concat([df_net_value["net_value_usd"], df_outstanding_shares["outstanding_shares"]], axis=1)

            # we reindex to account for missing pos.
            df = df.reindex(
                pd.MultiIndex.from_product(
                    [
                        df.index.levels[0],
                        pd.date_range(df.index.levels[1].min(), df.index.levels[1].max(), freq="W-MON"),
                    ],  # we downsample to make the chart easier on the frontend
                    names=["instrument", "date"],
                )
            )

            # forward fill around the instrument
            df = df.groupby(level=0).ffill().reset_index()

            df["market_capitalization_usd"] = df["outstanding_shares"] * df["net_value_usd"]

            df["instrument_repr"] = df["instrument"].map(dict(Product.objects.values_list("id", "computed_str")))
            df["instrument"] = df["instrument"].map(dict(Product.objects.values_list("id", "isin")))
            df = df.where(pd.notnull(df), None)
            fig = px.area(
                df,
                x="date",
                y="market_capitalization_usd",
                line_group="instrument",
                hover_name="instrument_repr",
                color="instrument",
            )
        return fig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                instrument__in=Product.objects.values("id"),
            )
            .select_related("instrument")
        )


@receiver(add_instance_button, sender=ProductModelViewSet)
def add_product_instrument_request_button(sender, many, *args, request=None, view=None, pk=None, **kwargs):
    return InstrumentButtonMixin.add_instrument_request_button(request=request, view=view, pk=pk)


@receiver(add_instance_button, sender=ProductModelViewSet)
def add_product_transactions_request_button(sender, many, *args, request=None, view=None, pk=None, **kwargs):
    return InstrumentButtonMixin.add_transactions_request_button(request=request, view=view, pk=pk)



from typing import TYPE_CHECKING, Any

from django.db.models import F, FloatField, Sum
from django.db.models.functions import Cast, ExtractYear
from langchain_core.messages import HumanMessage, SystemMessage
from wbfdm.models import Instrument

if TYPE_CHECKING:
    from langchain_core.messages import BaseMessage
    from wbcrm.models import Account


def get_holding_prompt(account: "Account") -> tuple[list["BaseMessage"], dict[str, Any]]:
    from wbportfolio.models import Product
    from wbportfolio.models.transactions.claim import Claim

    products = (
        Claim.get_valid_and_approved_claims(account=account)
        .distinct("product")
        .values_list("product", "product__isin")
    )

    performances = {}
    for product_id, product_name in products:
        performances[product_name] = (
            Instrument.extract_annual_performance_df(Product.objects.get(id=product_id).get_prices_df())
            .set_index("year")["performance"]
            .to_dict()
        )

    return [
        SystemMessage(
            "The following products are held by the account holder. Analyze their performances and check correlations between the holdings and their performances/interactions."
        ),
        HumanMessage("Performances per product ISIN per year: {performances}"),
    ], {"performances": performances}


def get_performances_prompt(account: "Account") -> tuple[list["BaseMessage"], dict[str, Any]]:
    from wbportfolio.models.transactions.claim import Claim

    holdings = (
        Claim.get_valid_and_approved_claims(account=account)
        .annotate(year=ExtractYear("date"))
        .values("year", "product")
        .annotate(
            sum_shares=Cast(Sum("shares"), FloatField()),
            product_name=F("product__name"),
            product_isin=F("product__isin"),
        )
        .values("product_name", "product_isin", "sum_shares", "year")
    ).order_by("year", "product")

    return [
        SystemMessage(
            "The following holdings (subscriptions/redemptions) have been found for this account. Please include this data in the analysis and check if there is any correlation between the holding data and the interactions."
        ),
        HumanMessage("List of holdings for the account: {holdings}"),
    ], {"holdings": list(holdings)}



# Generated by Django 5.0.14 on 2025-07-16 12:39

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0080_alter_trade_drift_factor_alter_trade_weighting'),
    ]

    operations = [
        migrations.AlterField(
            model_name='trade',
            name='drift_factor',
            field=models.DecimalField(decimal_places=16, default=Decimal('1'), help_text='Drift factor to be applied to the previous portfolio weight to get the actual effective weight including daily return', max_digits=19, verbose_name='Drift Factor'),
        ),
    ]



import logging
from collections import defaultdict
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Generator, Iterable

import numpy as np
import pandas as pd
from celery import shared_task
from django.contrib.postgres.fields import DateRangeField
from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import Exists, F, OuterRef, Q, QuerySet, Sum
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from pandas._libs.tslibs.offsets import BDay
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.models import WBModel
from wbcore.utils.models import ActiveObjectManager, DeleteToDisableMixin
from wbcore.workers import Queue
from wbfdm.models import Cash, Instrument, InstrumentType
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbfdm.signals import investable_universe_updated

from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.builder import AssetPositionBuilder
from wbportfolio.models.indexes import Index
from wbportfolio.models.portfolio_relationship import (
    InstrumentPortfolioThroughModel,
    PortfolioInstrumentPreferredClassificationThroughModel,
)
from wbportfolio.models.products import Product
from wbportfolio.pms.analytics.portfolio import Portfolio as AnalyticPortfolio
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position as PositionDTO

from ..constants import EQUITY_TYPE_KEYS
from ..order_routing.adapters import BaseCustodianAdapter
from . import PortfolioRole, ProductGroup

logger = logging.getLogger("pms")
if TYPE_CHECKING:
    from wbportfolio.models.orders.order_proposals import OrderProposal

MARKET_HOLIDAY_MAX_DURATION = 15


class DefaultPortfolioQueryset(QuerySet):
    def filter_invested_at_date(self, val_date: date) -> QuerySet:
        """
        Filter the queryset to get only portfolio invested at the given date
        """
        return self.filter(
            (Q(invested_timespan__startswith__lte=val_date) | Q(invested_timespan__startswith__isnull=True))
            & (Q(invested_timespan__endswith__gt=val_date) | Q(invested_timespan__endswith__isnull=True))
        )

    def filter_active_and_tracked(self):
        return self.annotate(
            has_product=Exists(
                InstrumentPortfolioThroughModel.objects.filter(
                    instrument__instrument_type=InstrumentType.PRODUCT, portfolio=OuterRef("pk")
                )
            )
        ).filter((Q(has_product=True) | Q(is_manageable=True)) & Q(is_active=True) & Q(is_tracked=True))

    def to_dependency_iterator(self, val_date: date) -> Iterable["Portfolio"]:
        """
        A method to sort the given queryset to return undependable portfolio first. This is very useful if a routine needs to be applied sequentially on portfolios by order of dependence.
        """
        max_iterations: int = (
            5  # in order to avoid circular dependency and infinite loop, we need to stop recursion at a max depth
        )
        remaining_portfolios = set(self)

        def _iterator(p, iterator_counter=0):
            iterator_counter += 1
            parent_portfolios = remaining_portfolios & set(
                map(lambda o: o[0], p.get_parent_portfolios(val_date))
            )  # get composition parent portfolios
            dependency_relationships = PortfolioPortfolioThroughModel.objects.filter(
                portfolio=p, dependency_portfolio__in=remaining_portfolios
            )  # get dependency portfolios
            if iterator_counter >= max_iterations or (
                not dependency_relationships.exists() and not bool(parent_portfolios)
            ):  # if not dependency portfolio or parent portfolio that remained, then we yield
                remaining_portfolios.remove(p)
                yield p
            else:
                # otherwise, we iterate of the dependency portfolio first
                deps_portfolios = parent_portfolios.union(
                    set([r.dependency_portfolio for r in dependency_relationships])
                )
                for deps_p in deps_portfolios:
                    yield from _iterator(deps_p, iterator_counter=iterator_counter)

        while len(remaining_portfolios) > 0:
            portfolio = next(iter(remaining_portfolios))
            yield from _iterator(portfolio)


class DefaultPortfolioManager(ActiveObjectManager):
    def get_queryset(self):
        return DefaultPortfolioQueryset(self.model).filter(is_active=True)

    def filter_invested_at_date(self, val_date: date):
        return self.get_queryset().filter_invested_at_date(val_date)

    def filter_active_and_tracked(self):
        return self.get_queryset().filter_active_and_tracked()


class ActiveTrackedPortfolioManager(DefaultPortfolioManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(asset_exists=Exists(AssetPosition.unannotated_objects.filter(portfolio=OuterRef("pk"))))
            .filter(Q(asset_exists=True) & (Q(is_tracked=True) | Q(is_manageable=True)))
        )


class PortfolioPortfolioThroughModel(models.Model):
    class Type(models.TextChoices):
        LOOK_THROUGH = "LOOK_THROUGH", "Look-through"
        MODEL = "MODEL", "Model"
        CUSTODIAN = "CUSTODIAN", "Custodian"
        HIERARCHICAL = "HIERARCHICAL", "Hierarchical"

    portfolio = models.ForeignKey("wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependency_through")
    dependency_portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependent_through"
    )
    type = models.CharField(choices=Type.choices, default=Type.LOOK_THROUGH, verbose_name="Type")

    def __str__(self):
        return f"{self.portfolio} dependant on {self.dependency_portfolio} ({self.Type[self.type].label})"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["portfolio", "type"], name="unique_lookthrough", condition=Q(type="LOOK_THROUGH")
            ),
            models.UniqueConstraint(fields=["portfolio", "type"], name="unique_model", condition=Q(type="MODEL")),
        ]


class Portfolio(DeleteToDisableMixin, WBModel):
    assets: models.QuerySet[AssetPosition]
    builder: AssetPositionBuilder

    name = models.CharField(
        max_length=255,
        verbose_name="Name",
        default="",
        help_text="The Name of the Portfolio",
    )

    currency = models.ForeignKey(
        to="currency.Currency",
        related_name="portfolios",
        on_delete=models.PROTECT,
        verbose_name="Currency",
        help_text="The currency of the portfolio.",
    )
    hedged_currency = models.ForeignKey(
        to="currency.Currency",
        related_name="hedged_portfolios",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name="Hedged Currency",
        help_text="The hedged currency of the portfolio.",
    )
    depends_on = models.ManyToManyField(
        "wbportfolio.Portfolio",
        symmetrical=False,
        related_name="dependent_portfolios",
        through="wbportfolio.PortfolioPortfolioThroughModel",
        through_fields=("portfolio", "dependency_portfolio"),
        blank=True,
        verbose_name="The portfolios this portfolio depends on",
    )

    preferred_instrument_classifications = models.ManyToManyField(
        "wbfdm.Instrument",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
        related_name="preferred_portfolio_classifications",
        through="wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel",
        through_fields=("portfolio", "instrument"),
        blank=True,
        verbose_name="The Preferred classification per instrument",
    )
    instruments = models.ManyToManyField(
        "wbfdm.Instrument",
        through=InstrumentPortfolioThroughModel,
        related_name="portfolios",
        blank=True,
        verbose_name="Instruments",
        help_text="Instruments linked to this instrument",
    )
    invested_timespan = DateRangeField(
        null=True, blank=True, help_text="Define when this portfolio is considered invested"
    )

    is_manageable = models.BooleanField(
        default=False,
        help_text="True if the portfolio can be manually modified (e.g. Order Proposal be submitted or total weight recomputed)",
    )
    is_tracked = models.BooleanField(
        default=True,
        help_text="True if the internal updating mechanism (e.g., Next weights or Look-Through computation, rebalancing etc...) needs to apply to this portfolio",
    )
    is_lookthrough = models.BooleanField(
        default=False,
        help_text="Indicates that this portfolio is a look-through portfolio",
    )
    is_composition = models.BooleanField(
        default=False, help_text="If true, this portfolio is a composition of other portfolio"
    )
    only_weighting = models.BooleanField(
        default=False,
        help_text="Indicates that this portfolio is only utilizing weights and disregards shares, e.g. a model portfolio",
    )
    only_keep_essential_positions = models.BooleanField(
        default=False, help_text="Only keep essential positions when drifting positions"
    )
    updated_at = models.DateTimeField(blank=True, null=True, verbose_name="Updated At")
    last_position_date = models.DateField(blank=True, null=True, verbose_name="Last Position Date")
    initial_position_date = models.DateField(blank=True, null=True, verbose_name="Last Position Date")

    bank_accounts = models.ManyToManyField(
        to="directory.BankingContact",
        related_name="wbportfolio_portfolios",
        through="wbportfolio.PortfolioBankAccountThroughModel",
        blank=True,
    )

    # OMS default parameters. Used to seed order proposal default value upon creation
    default_order_proposal_min_order_value = models.IntegerField(
        default=0, verbose_name="Default Order Proposal Minimum Order Value"
    )
    default_order_proposal_min_weighting = models.DecimalField(
        max_digits=9,
        decimal_places=8,
        default=Decimal(0),
        verbose_name="Default Order Proposal Minimum Weight",
        validators=[MinValueValidator(Decimal("0")), MaxValueValidator(Decimal("1"))],
    )
    default_order_proposal_total_cash_weight = models.DecimalField(
        default=Decimal("0"),
        decimal_places=4,
        max_digits=5,
        verbose_name="Default Order Proposal Total Cash Weight",
        validators=[MinValueValidator(Decimal("0")), MaxValueValidator(Decimal("1"))],
    )

    objects = DefaultPortfolioManager()
    tracked_objects = ActiveTrackedPortfolioManager()

    def __init__(self, *args, **kwargs):
        self.builder = AssetPositionBuilder(self)
        super().__init__(*args, **kwargs)

    @property
    def primary_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.LOOK_THROUGH
            ).dependency_portfolio

    @property
    def model_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
            ).dependency_portfolio

    @property
    def composition_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self,
                type=PortfolioPortfolioThroughModel.Type.MODEL,
                dependency_portfolio__is_composition=True,
            ).dependency_portfolio

    @property
    def is_model(self) -> bool:
        return PortfolioPortfolioThroughModel.objects.filter(
            type=PortfolioPortfolioThroughModel.Type.MODEL,
            dependency_portfolio=self,
        ).exists()

    @property
    def imported_assets(self):
        return self.assets.filter(is_estimated=False)

    @cached_property
    def pms_instruments(self):
        instruments = [i for i in Product.objects.filter(portfolios=self)]
        instruments.extend([i for i in ProductGroup.objects.filter(portfolios=self)])
        instruments.extend([i for i in Index.objects.filter(portfolios=self)])
        return instruments

    @property
    def cash_component(self) -> Cash:
        return Cash.objects.get_or_create(
            currency=self.currency, defaults={"is_cash": True, "name": self.currency.title}
        )[0]

    def get_authenticated_custodian_adapter(self, **kwargs) -> BaseCustodianAdapter | None:
        supported_instruments_for_routing = list(
            filter(lambda o: o.order_routing_custodian_adapter, self.pms_instruments)
        )
        if not supported_instruments_for_routing:
            raise ValueError("No custodian adapter for this portfolio")

        pms_instrument = supported_instruments_for_routing[
            0
        ]  # for simplicity we support only one instrument per portfolio that is allowed to support order routing
        adapter = import_string(pms_instrument.order_routing_custodian_adapter)(
            isin=pms_instrument.isin, identifier=pms_instrument.identifier, **kwargs
        )
        adapter.authenticate()
        if not adapter.is_valid():
            raise ValueError("This portfolio is not valid for rebalancing")
        return adapter

    @property
    def can_be_rebalanced(self):
        return self.is_manageable and not self.is_lookthrough

    def delete(self, **kwargs):
        super().delete(**kwargs)
        # We check if for all linked instruments, this portfolio was the last active one (if yes, we disable the instrument)
        if self.id:
            for instrument in self.instruments.iterator():
                if not instrument.portfolios.filter(is_active=True).exists():
                    instrument.delisted_date = date.today() - timedelta(days=1)
                    instrument.save()

    def _build_dto(self, val_date: date, **extra_kwargs) -> PortfolioDTO:
        "returns the dto representation of this portfolio at the specified date"
        assets = self.assets.filter(date=val_date, **extra_kwargs)
        try:
            last_returns, _ = self.get_analytic_portfolio(val_date, use_dl=True).get_contributions()
            last_returns = last_returns.to_dict()
        except ValueError:
            last_returns = {}
        positions = []
        for asset in assets:
            positions.append(asset._build_dto(daily_return=last_returns.get(asset.underlying_quote.id, Decimal("0"))))

        return PortfolioDTO(positions)

    def get_weights(self, val_date: date) -> dict[int, float]:
        """
        A convenience utility method to returns the portfolio weights for this portfolio as a dictionary (instrument id as key and weights as value)

        Args:
            val_date: The date at which to return the weights for this portfolio

        Returns:
            A dictionary containing the weights for this portfolio
        """
        return dict(
            map(
                lambda r: (r[0], float(r[1])),
                self.assets.filter(date=val_date)
                .values("underlying_quote")
                .annotate(sum_weight=Sum("weighting"))
                .values_list("underlying_quote", "sum_weight"),
            )
        )

    def get_analytic_portfolio(
        self, val_date: date, weights: dict[int, float] | None = None, use_dl: bool = True, **kwargs
    ) -> AnalyticPortfolio:
        """
        Return the analytic portfolio associated with this portfolio at the given date

        the analytic portfolio inherit from SKFolio Portfolio and can be used to access all this library methods
        Args:
            val_date: the date to calculate the portfolio for

        Returns:
            The instantiated analytic portfolio
        """
        if not weights:
            weights = self.get_weights(val_date)
        return_date = (val_date + BDay(1)).date()
        returns = self.load_builder_returns(val_date, return_date, use_dl=use_dl).copy()
        if pd.Timestamp(return_date) not in returns.index:
            raise ValueError()
        returns = returns.loc[:return_date, :]
        returns = returns.fillna(0)  # not sure this is what we want
        return AnalyticPortfolio(
            X=returns,
            weights=weights,
        )

    def is_invested_at_date(self, val_date: date) -> bool:
        return (
            self.invested_timespan
            and self.invested_timespan.upper > val_date
            and self.invested_timespan.lower <= val_date
        )

    def __str__(self):
        return f"{self.id:06}: {self.name}"

    class Meta:
        verbose_name = "Portfolio"
        verbose_name_plural = "Portfolios"

        notification_types = [
            create_notification_type(
                "wbportfolio.portfolio.check_custodian_portfolio",
                "Check Custodian Portfolio",
                "Sends a notification when a portfolio does not match with its custodian portfolio",
                True,
                True,
                True,
            ),
            create_notification_type(
                "wbportfolio.portfolio.action_done",
                "Portfolio Action finished",
                "Sends a notification when a the requested portfolio action is done (e.g. replay, quote adjustment...)",
                True,
                True,
                True,
                is_lock=True,
            ),
            create_notification_type(
                "wbportfolio.portfolio.warning",
                "PMS Warning",
                "Sends a notification to warn portfolio manager or administrator regarding issue that needs action.",
                True,
                True,
                True,
                is_lock=True,
            ),
        ]

    def is_active_at_date(self, val_date: date) -> bool:
        """
        Return if the base instrument has a total aum greater than 0
        :val_date: the date at which we need to evaluate if the portfolio is considered active
        """
        active_portfolio = self.is_active or self.deletion_datetime.date() > val_date
        if self.instruments.exists():
            return active_portfolio and any(
                [instrument.is_active_at_date(val_date) for instrument in self.instruments.all()]
            )
        return active_portfolio

    def get_total_asset_value(self, val_date: date) -> Decimal:
        """
        Return the total asset under management of the portfolio at the specified valuation date
        Args:
            val_date: The date at which aum needs to be computed
        Returns:
            The total AUM (0 if there is no position)
        """
        return self.assets.filter(date=val_date).aggregate(s=Sum("total_value_fx_portfolio"))["s"] or Decimal(0.0)

    def get_total_asset_under_management(self, val_date):
        from wbportfolio.models.transactions.trades import Trade

        trades = Trade.valid_customer_trade_objects.filter(portfolio=self, transaction_date__lte=val_date)

        total_aum = Decimal(0)
        for underlying_instrument_id, sum_shares in (
            trades.values("underlying_instrument")
            .annotate(
                sum_shares=Sum("shares"),
            )
            .values_list("underlying_instrument", "sum_shares")
        ):
            with suppress(Instrument.DoesNotExist, InstrumentPrice.DoesNotExist):
                instrument = Instrument.objects.get(id=underlying_instrument_id)
                last_price = instrument.valuations.filter(date__lte=val_date).latest("date").net_value
                fx_rate = instrument.currency.convert(val_date, self.currency)
                total_aum += last_price * sum_shares * fx_rate
        return total_aum

    def _get_assets(self, with_estimated=True, with_cash=True):
        qs = self.assets
        if not with_estimated:
            qs = qs.filter(is_estimated=False)
        if not with_cash:
            qs = qs.exclude(underlying_instrument__is_cash=True)
        return qs

    def get_earliest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__gte=val_date)
        if qs.exists():
            return qs.earliest("date").date
        return None

    def get_latest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__lte=val_date)

        if qs.exists():
            return qs.latest("date").date
        return None

    # Asset Position Utility Functions
    def get_holding(self, val_date, exclude_cash=True, exclude_index=True):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date, weighting__gt=0)
        if exclude_index:
            qs = qs.exclude(underlying_instrument__instrument_type=InstrumentType.INDEX)
        return (
            qs.values("underlying_instrument__name")
            .annotate(total_value_fx_portfolio=Sum("total_value_fx_portfolio"), weighting=Sum("weighting"))
            .order_by("-total_value_fx_portfolio")
        )

    def _get_groupedby_df(
        self,
        group_by,
        val_date: date,
        exclude_cash: bool | None = False,
        exclude_index: bool | None = False,
        extra_filter_parameters: dict[str, Any] = None,
        **groupby_kwargs,
    ):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date)
        if exclude_index:
            # We exclude only index that are not considered as cash. Setting exclude_cash to true convers this case.
            qs = qs.exclude(
                Q(underlying_instrument__instrument_type=InstrumentType.INDEX)
                & Q(underlying_instrument__is_cash=False)
            )
        if extra_filter_parameters:
            qs = qs.filter(**extra_filter_parameters)
        qs = group_by(qs, **groupby_kwargs).annotate(sum_weighting=Sum(F("weighting"))).order_by("-sum_weighting")
        df = pd.DataFrame(
            qs.values_list("aggregated_title", "sum_weighting"), columns=["aggregated_title", "weighting"]
        )
        if not df.empty:
            df.weighting = df.weighting.astype("float")
            df.weighting = df.weighting / df.weighting.sum()
            df = df.sort_values(by=["weighting"])
        return df.where(pd.notnull(df), None)

    def get_geographical_breakdown(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.country_group_by, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_currency_exposure(self, val_date, **kwargs):
        df = self._get_groupedby_df(AssetPosition.currency_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_market_cap_distribution(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.marketcap_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_instrument__instrument_type__key__in": EQUITY_TYPE_KEYS},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_liquidity(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.liquidity_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_instrument__instrument_type__key__in": EQUITY_TYPE_KEYS},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_industry_exposure(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.group_by_primary, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_asset_allocation(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(AssetPosition.cash_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_adjusted_child_positions(self, val_date):
        if (
            child_positions := self.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        ).count() == 1:
            if portfolio := child_positions.first().underlying_instrument.primary_portfolio:
                child_positions = portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        for position in child_positions:
            if child_portfolio := position.underlying_instrument.primary_portfolio:
                index_positions = child_portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(
                    date=val_date
                )

                for index_position in index_positions.all():
                    weighting = index_position.weighting * position.weighting
                    if weighting != 0:
                        yield {
                            "underlying_instrument_id": index_position.underlying_instrument.id,
                            "weighting": weighting,
                        }

    def get_longshort_distribution(self, val_date):
        df = pd.DataFrame(self.get_adjusted_child_positions(val_date))

        if not df.empty:
            df["is_cash"] = df.underlying_instrument_id.apply(lambda x: Instrument.objects.get(id=x).is_cash)
            df = df[~df["is_cash"]]
            df = (
                df[["underlying_instrument_id", "weighting"]].groupby("underlying_instrument_id").sum().astype("float")
            )
            df.weighting = df.weighting / df.weighting.sum()
            short_weight = df[df.weighting < 0].weighting.abs().sum()
            long_weight = df[df.weighting > 0].weighting.sum()
            total_weight = long_weight + short_weight
            return pd.DataFrame(
                [
                    {"title": "Long", "weighting": long_weight / total_weight},
                    {"title": "Short", "weighting": short_weight / total_weight},
                ]
            )
        return df

    def get_portfolio_contribution_df(
        self,
        start: date,
        end: date,
        with_cash: bool = True,
        hedged_currency: Currency | None = None,
        only_equity: bool = False,
    ) -> pd.DataFrame:
        qs = self._get_assets(with_cash=with_cash).filter(date__gte=start, date__lte=end)
        if only_equity:
            qs = qs.filter(underlying_instrument__instrument_type__key__in=EQUITY_TYPE_KEYS)
        qs = qs.annotate_hedged_currency_fx_rate(hedged_currency)
        df = Portfolio.get_contribution_df(
            qs.select_related("underlying_instrument").values_list(
                "date", "price", "hedged_currency_fx_rate", "underlying_instrument", "weighting"
            )
        )
        df = df.rename(columns={"group_key": "underlying_instrument"})
        df["underlying_instrument__name_repr"] = df["underlying_instrument"].map(
            dict(Instrument.objects.filter(id__in=df["underlying_instrument"]).values_list("id", "name_repr"))
        )
        return df

    def check_related_portfolio_at_date(self, val_date: date, related_portfolio: "Portfolio"):
        assets = AssetPosition.objects.filter(
            date=val_date, underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ).values("underlying_instrument", "shares")
        assets1 = assets.filter(portfolio=self)
        assets2 = assets.filter(portfolio=related_portfolio)
        return assets1.difference(assets2)

    def get_child_portfolios(self, val_date: date) -> set["Portfolio"]:
        child_portfolios = set()
        instrument_rel = InstrumentPortfolioThroughModel.objects.filter(portfolio=self)
        if instrument_rel.exists():
            for parent_portfolio in Portfolio.objects.filter(
                id__in=AssetPosition.unannotated_objects.filter(
                    date=val_date, underlying_quote__in=instrument_rel.values("instrument")
                ).values("portfolio")
            ):
                child_portfolios.add(parent_portfolio)
        return child_portfolios

    def get_parent_portfolios(self, val_date: date) -> set["Portfolio"]:
        for asset in self.assets.filter(date=val_date, underlying_instrument__portfolios__isnull=False).distinct(
            "underlying_instrument"
        ):
            if portfolio := asset.underlying_instrument.portfolio:
                yield portfolio, asset.weighting

    def get_next_rebalancing_date(self, start_date: date) -> date | None:
        if automatic_rebalancer := getattr(self, "automatic_rebalancer", None):
            return automatic_rebalancer.get_next_rebalancing_date(start_date)

    def fix_quantization(self, val_date: date):
        assets = self.assets.filter(date=val_date)
        total_weighting = assets.aggregate(s=Sum("weighting"))["s"]
        if total_weighting and (quantization_error := Decimal("1") - total_weighting):
            cash = self.cash_component
            try:
                cash_pos = assets.get(underlying_quote=cash)
                cash_pos.weighting += quantization_error
            except AssetPosition.DoesNotExist:
                cash_pos = AssetPosition(
                    portfolio=self,
                    underlying_quote=cash,
                    weighting=quantization_error,
                    initial_price=Decimal("1"),
                    date=val_date,
                    is_estimated=True,
                )
            cash_pos.save(create_underlying_quote_price_if_missing=True)

    def change_at_date(
        self,
        val_date: date,
        fix_quantization: bool = False,
        evaluate_rebalancer: bool = True,
        changed_portfolio: AnalyticPortfolio | None = None,
        broadcast_changes_at_date: bool = True,
        **kwargs,
    ):
        if not self.is_tracked:
            return
        logger.info(f"change at date for {self} at {val_date}")

        if fix_quantization:
            # We assume all ptf total weight is 100% but quantization error can occur. In that case, we create a cash component and add the weight there.
            self.fix_quantization(val_date)

        # We check if there is an instrument attached to the portfolio with calculated NAV and price computation method
        self.estimate_net_asset_values(
            (val_date + BDay(1)).date(), analytic_portfolio=changed_portfolio
        )  # updating weighting in t0 influence nav in t+1
        if evaluate_rebalancer:
            self.evaluate_rebalancing(val_date)

        self.updated_at = timezone.now()
        if self.assets.filter(date=val_date).exists():
            if not self.last_position_date or self.last_position_date < val_date:
                self.last_position_date = val_date
            if not self.initial_position_date or self.initial_position_date > val_date:
                self.initial_position_date = val_date
        self.save()
        if broadcast_changes_at_date:
            self.handle_controlling_portfolio_change_at_date(
                val_date,
                fix_quantization=fix_quantization,
                changed_portfolio=changed_portfolio,
                **kwargs,
            )

    def handle_controlling_portfolio_change_at_date(self, val_date: date, **kwargs):
        if self.is_tracked:
            for rel in PortfolioPortfolioThroughModel.objects.filter(
                dependency_portfolio=self,
                type=PortfolioPortfolioThroughModel.Type.LOOK_THROUGH,
                portfolio__is_lookthrough=True,
            ):
                rel.portfolio.compute_lookthrough(val_date)
            for rel in PortfolioPortfolioThroughModel.objects.filter(
                dependency_portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
            ):
                rel.portfolio.evaluate_rebalancing(val_date)
            for dependent_portfolio in self.get_child_portfolios(val_date):
                # dependent_portfolio.change_at_date(val_date, **kwargs)
                dependent_portfolio.handle_controlling_portfolio_change_at_date(val_date, **kwargs)

    def get_model_portfolio_relationships(
        self, val_date: date
    ) -> Generator[PortfolioPortfolioThroughModel, None, None]:
        for rel in PortfolioPortfolioThroughModel.objects.filter(
            dependency_portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
        ):
            if rel.portfolio.is_active_at_date(val_date):
                yield rel
        for dependent_portfolio in self.get_child_portfolios(val_date):
            yield from dependent_portfolio.get_model_portfolio_relationships(val_date)

    def evaluate_rebalancing(self, val_date: date):
        if hasattr(self, "automatic_rebalancer"):
            # if the portfolio has an automatic rebalancer and the next business day is suitable with the rebalancer, we create a order proposal automatically
            next_business_date = (val_date + BDay(1)).date()
            if self.automatic_rebalancer.is_valid(val_date):  # we evaluate the rebalancer in t0 and t+1
                logger.info(f"Evaluate Rebalancing for {self} at {val_date}")
                self.automatic_rebalancer.evaluate_rebalancing(val_date)
            if self.automatic_rebalancer.is_valid(next_business_date):
                logger.info(f"Evaluate Rebalancing for {self} at {next_business_date}")
                self.automatic_rebalancer.evaluate_rebalancing(next_business_date)

    def estimate_net_asset_values(self, val_date: date, analytic_portfolio: AnalyticPortfolio | None = None):
        effective_portfolio_date = (val_date - BDay(1)).date()
        with suppress(ValueError):
            if not analytic_portfolio:
                analytic_portfolio = self.get_analytic_portfolio(effective_portfolio_date, use_dl=False)
            for instrument in self.pms_instruments:
                # we assume that in t-1 we will have a portfolio (with at least estimate position). If we use the latest position date before val_date, we run into the problem of being able to compute nav at every date
                last_price = instrument.get_latest_price(effective_portfolio_date)
                if (
                    instrument.is_active_at_date(val_date)
                    and (net_asset_value_computation_method_path := instrument.net_asset_value_computation_method_path)
                    and last_price
                ):
                    logger.info(f"Estimate NAV of {val_date:%Y-%m-%d} for instrument {instrument}")
                    net_asset_value_computation_method = import_string(net_asset_value_computation_method_path)
                    estimated_net_asset_value = net_asset_value_computation_method(last_price, analytic_portfolio)
                    if estimated_net_asset_value is not None:
                        InstrumentPrice.objects.update_or_create(
                            instrument=instrument,
                            date=val_date,
                            calculated=True,
                            defaults={
                                "gross_value": estimated_net_asset_value,
                                "net_value": estimated_net_asset_value,
                            },
                        )
                        if (
                            val_date == instrument.last_price_date
                        ):  # if price date is the latest instrument price date, we recompute the last valuation data
                            instrument.update_last_valuation_date()

    def drift_weights(
        self, start_date: date, end_date: date, stop_at_rebalancing: bool = False
    ) -> Generator[tuple[date, dict[int, float]], None, models.Model]:
        logger.info(f"drift weights for {self} from {start_date:%Y-%m-%d} to {end_date:%Y-%m-%d}")

        rebalancer = getattr(self, "automatic_rebalancer", None)
        # Get initial weights
        weights = self.get_weights(start_date)  # initial weights
        if not weights:
            previous_date = self.assets.filter(date__lte=start_date).latest("date").date
            _, weights = next(self.drift_weights(previous_date, start_date))
        last_order_proposal = None
        for to_date_ts in pd.date_range(start_date + timedelta(days=1), end_date, freq="B"):
            to_date = to_date_ts.date()
            to_is_active = self.is_active_at_date(to_date)
            logger.info(f"Processing {to_date:%Y-%m-%d}")
            order_proposal = None
            try:
                last_returns = self.builder.returns.loc[[to_date_ts], :]
                analytic_portfolio = AnalyticPortfolio(weights=weights, X=last_returns)
                drifted_weights = analytic_portfolio.get_next_weights()
            except KeyError:  # if no return for that date, we break and continue
                break
            try:
                order_proposal = self.order_proposals.get(
                    trade_date=to_date, rebalancing_model__isnull=True, status="CONFIRMED"
                )
            except ObjectDoesNotExist:
                if rebalancer and rebalancer.is_valid(to_date):
                    rebalancer.portfolio = self  # ensure reference is the same to access cached returns
                    effective_portfolio = PortfolioDTO(
                        positions=[
                            PositionDTO(
                                date=to_date,
                                underlying_instrument=i,
                                weighting=Decimal(w),
                                daily_return=Decimal(last_returns.iloc[-1][i]),
                            )
                            for i, w in weights.items()
                        ]
                    )
                    order_proposal = rebalancer.evaluate_rebalancing(to_date, effective_portfolio=effective_portfolio)
            if order_proposal:
                last_order_proposal = order_proposal
                if stop_at_rebalancing:
                    break
                next_weights = {
                    trade.underlying_instrument.id: float(trade._target_weight)
                    for trade in order_proposal.get_orders()
                }
                yield to_date, next_weights
            else:
                next_weights = drifted_weights
                if to_is_active:
                    yield to_date, next_weights
                else:
                    yield (
                        to_date,
                        {underlying_quote_id: 0.0 for underlying_quote_id in weights.keys()},
                    )  # if we have no return or portfolio is not active anymore, we return an emptied portfolio
                    break
            weights = next_weights
        return last_order_proposal

    def propagate_or_update_assets(self, from_date: date, to_date: date):
        """
        Create a new portfolio at `to_date` based on the portfolio in `from_date`.

        Args:
            from_date: The date to propagate the portfolio from
            to_date:  The date to create the new portfolio at

        """
        # we don't propagate on already imported portfolio by default
        is_target_portfolio_imported = self.assets.filter(date=to_date, is_estimated=False).exists()
        if (
            not self.is_lookthrough and not is_target_portfolio_imported and self.is_active_at_date(from_date)
        ):  # we cannot propagate a new portfolio for untracked, or look-through or already imported or inactive portfolios
            self.load_builder_returns(from_date, to_date)
            for pos_date, weights in self.drift_weights(from_date, to_date):
                self.builder.add((pos_date, weights))
            self.builder.bulk_create_positions(delete_leftovers=True)
            self.builder.schedule_change_at_dates()
            self.builder.schedule_metric_computation()

    def load_builder_returns(self, from_date: date, to_date: date, use_dl: bool = True) -> pd.DataFrame:
        instruments_ids = list(self.get_weights(from_date).keys())
        for tp in self.order_proposals.filter(trade_date__gte=from_date, trade_date__lte=to_date):
            instruments_ids.extend(tp.orders.values_list("underlying_instrument", flat=True))
        self.builder.load_returns(
            set(instruments_ids), (from_date - BDay(1)).date(), (to_date + BDay(1)).date(), use_dl=use_dl
        )
        return self.builder.returns

    def get_lookthrough_positions(
        self,
        sync_date: date,
        portfolio_total_asset_value: Decimal | None = None,
        with_intermediary_position: bool = False,
    ):
        """Recursively calculates the look-through position for a portfolio

        Arguments:
            sync_date {datetime.date} -- The date on which the assets will be computed
            portfolio_total_value: {Decimal} -- The total value of the portfolio (needed to compute initial shares)
        """

        def _crawl_portfolio(
            parent_portfolio,
            adjusted_weighting,
            adjusted_currency_fx_rate,
            adjusted_is_estimated,
            path=None,
        ):
            if not path:
                path = []
            path.append(parent_portfolio)
            for position in parent_portfolio.assets.filter(date=sync_date):
                position.id = None
                position.weighting = adjusted_weighting * position.weighting
                position.initial_currency_fx_rate = adjusted_currency_fx_rate * position.currency_fx_rate
                position.is_estimated = (adjusted_is_estimated or position.is_estimated) and not (
                    position.weighting == 1.0
                )
                # to get from which portfolio this position is created, we need to differantiate between:
                # * Composition portfolio: where the portfolio created is the second encountered portfolio
                # * Other: portfolio created is the last encountered portfolio
                # If `path` is empty, we use None as portfolio_created
                try:
                    if self.is_composition:
                        position.portfolio_created = path[1]
                    else:
                        position.portfolio_created = path[-1]
                except IndexError:
                    position.portfolio_created = None

                position.path = path
                position.initial_shares = None
                if portfolio_total_asset_value and (price_fx_portfolio := position.price * position.currency_fx_rate):
                    position.initial_shares = (position.weighting * portfolio_total_asset_value) / price_fx_portfolio
                if child_portfolio := position.underlying_quote.primary_portfolio:
                    if with_intermediary_position:
                        yield position
                    yield from _crawl_portfolio(
                        child_portfolio,
                        position.weighting,
                        position.currency_fx_rate,
                        position.is_estimated,
                        path=path.copy(),
                    )
                elif position.weighting:  # we do not yield position with weight 0 because of issue with certain multi-thematic portfolios which contain duplicates
                    yield position

        yield from _crawl_portfolio(self, Decimal(1.0), Decimal(1.0), False)

    def get_positions(self, val_date: date, **kwargs) -> Iterable[AssetPosition]:
        if self.is_composition:
            assets = list(self.get_lookthrough_positions(val_date, **kwargs))
        else:
            assets = list(self.assets.filter(date=val_date))
        return assets

    def compute_lookthrough(self, from_date: date, to_date: date | None = None):
        if not self.primary_portfolio or not self.is_lookthrough:
            raise ValueError(
                "Lookthrough position can only be computed on lookthrough portfolio with a primary portfolio"
            )
        if not to_date:
            to_date = from_date
        self.load_builder_returns(from_date, to_date)
        for val_date in pd.date_range(from_date, to_date, freq="B").date:
            logger.info(f"Compute Look-Through for {self} at {val_date}")
            portfolio_total_asset_value = (
                self.primary_portfolio.get_total_asset_under_management(val_date) if not self.only_weighting else None
            )
            self.builder.add(
                list(self.primary_portfolio.get_lookthrough_positions(val_date, portfolio_total_asset_value)),
                infer_underlying_quote_price=True,
            )
        self.builder.bulk_create_positions(delete_leftovers=True)
        self.builder.schedule_change_at_dates()
        self.builder.schedule_metric_computation()

    def update_preferred_classification_per_instrument(self):
        # Function to automatically assign Preferred instrument based on the assets' underlying instruments of the
        # attached wbportfolio
        instruments = filter(
            None,
            map(
                lambda x: Instrument.objects.get(id=x["underlying_instrument"]).get_classifable_ancestor(
                    include_self=True
                ),
                self.assets.values("underlying_instrument").distinct("underlying_instrument"),
            ),
        )
        leftovers_instruments = list(
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(portfolio=self).values_list(
                "instrument", flat=True
            )
        )
        for instrument in instruments:
            other_classifications = instrument.classifications.filter(group__is_primary=False)
            default_classification = None
            if other_classifications.count() == 1:
                default_classification = other_classifications.first()
            if not PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument
            ).exists():
                PortfolioInstrumentPreferredClassificationThroughModel.objects.create(
                    portfolio=self,
                    instrument=instrument,
                    classification=default_classification,
                    classification_group=default_classification.group if default_classification else None,
                )
            if instrument.id in leftovers_instruments:
                leftovers_instruments.remove(instrument.id)

        for instrument_id in leftovers_instruments:
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument_id
            ).delete()

    def reset_portfolio_with_rebalancing_model(self, inception_date: date) -> "OrderProposal":
        """
        WARNING: Use with caution!
        """
        if not self.is_manageable:
            raise ValueError("You cannot reset a portfolio that is not manageable")

        # ensure the attached rebalancer as the proper new activation date and activate automatically
        rebalancer = self.automatic_rebalancer
        rebalancer.activation_date = inception_date
        rebalancer.apply_order_proposal_automatically = True
        rebalancer.save()

        # for each ptf index, erase the existing prices and reset its inception date
        for index in self.instruments.all():
            index.prices.all().delete()
            index.inception_date = inception_date
            index.save()

        # reset the portfolio composition (expect to be recomputed during replay)
        self.assets.all().delete()

        # for each existing orders, arise all (expect to be recomputed with the rebalancer during replay)
        for op in self.order_proposals.all():
            op.orders.all().delete()
            op.delete()

        # create the seeding order proposal and refresh its orders
        op = self.order_proposals.create(
            trade_date=inception_date, comment="Initial Orders", rebalancing_model=rebalancer.rebalancing_model
        )
        op.reset_orders()
        return op

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfolio"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:portfoliorepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    @classmethod
    def _get_or_create_portfolio(cls, instrument_handler, portfolio_data):
        if isinstance(portfolio_data, int):
            return Portfolio.all_objects.get(id=portfolio_data)
        instrument = portfolio_data
        if isinstance(portfolio_data, dict):
            instrument = instrument_handler.process_object(instrument, only_security=False, read_only=True)[0]
        return instrument.primary_portfolio

    def check_share_diff(self, val_date: date) -> bool:
        return self.assets.filter(Q(date=val_date) & ~Q(initial_shares=F("initial_shares_at_custodian"))).exists()

    @classmethod
    def get_contribution_df(cls, data, need_normalize: bool = False):
        df = pd.DataFrame(
            data,
            columns=[
                "date",
                "price",
                "currency_fx_rate",
                "group_key",
                "value",
            ],
        )
        if not df.empty:
            df = df[df["value"] != 0]
            df.date = pd.to_datetime(df.date)
            df["price_fx_portfolio"] = df.price * df.currency_fx_rate

            df[["price", "price_fx_portfolio", "value", "currency_fx_rate"]] = df[
                ["price", "price_fx_portfolio", "value", "currency_fx_rate"]
            ].astype("float")

            df["group_key"] = df["group_key"].fillna(0)

            df = (
                df[
                    [
                        "group_key",
                        "date",
                        "price",
                        "price_fx_portfolio",
                        "value",
                        "currency_fx_rate",
                    ]
                ]
                .groupby(["date", "group_key"], dropna=False)
                .agg(
                    {
                        "price": "mean",
                        "price_fx_portfolio": "mean",
                        "value": "sum",
                        "currency_fx_rate": "mean",
                    }
                )
                .reset_index()
                .set_index("date")
                .sort_index()
            )
            df["value"] = df["value"].fillna(0)
            value = df.pivot_table(
                index="date",
                columns=["group_key"],
                values="value",
                fill_value=0,
                aggfunc="sum",
            )
            weights_ = value
            if need_normalize:
                total_value_price = df["value"].groupby("date", dropna=False).sum()
                weights_ = value.divide(total_value_price, axis=0)
            prices_usd = (
                df.pivot_table(
                    index="date",
                    columns=["group_key"],
                    values="price_fx_portfolio",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            rates_fx = (
                df.pivot_table(
                    index="date",
                    columns=["group_key"],
                    values="currency_fx_rate",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            prices_usd = prices_usd.ffill()
            performance_prices = prices_usd / prices_usd.shift(1, axis=0) - 1
            contributions_prices = performance_prices.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_prices = (1 + contributions_prices.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_prices = contributions_prices.multiply(total_contrib_prices, axis=0).sum(skipna=False)
            monthly_perf_prices = (1 + performance_prices).dropna(how="all").product(axis=0, skipna=False) - 1

            rates_fx = rates_fx.ffill()
            performance_rates_fx = rates_fx / rates_fx.shift(1, axis=0) - 1
            contributions_rates_fx = performance_rates_fx.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_rates_fx = (1 + contributions_rates_fx.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_rates_fx = contributions_rates_fx.multiply(total_contrib_rates_fx, axis=0).sum(skipna=False)
            monthly_perf_rates_fx = (1 + performance_rates_fx).dropna(how="all").product(axis=0, skipna=False) - 1

            res = pd.concat(
                [
                    monthly_perf_prices,
                    monthly_perf_rates_fx,
                    contributions_prices,
                    contributions_rates_fx,
                    weights_.iloc[0, :],
                    weights_.iloc[-1, :],
                    value.iloc[0, :],
                    value.iloc[-1, :],
                ],
                axis=1,
            ).reset_index()
            res.columns = [
                "group_key",
                "performance_total",
                "performance_forex",
                "contribution_total",
                "contribution_forex",
                "allocation_start",
                "allocation_end",
                "total_value_start",
                "total_value_end",
            ]

            return res.replace([np.inf, -np.inf, np.nan], 0)
        return pd.DataFrame()

    def get_or_create_index(self):
        index = Index.objects.get_or_create(name=self.name, defaults={"currency": self.currency})[0]
        index.portfolios.all().delete()
        InstrumentPortfolioThroughModel.objects.update_or_create(instrument=index, defaults={"portfolio": self})
        return index

    @classmethod
    def create_model_portfolio(cls, name: str, currency: Currency, with_index: bool = True):
        portfolio = cls.objects.create(
            is_manageable=True,
            name=name,
            currency=currency,
        )
        if with_index:
            portfolio.get_or_create_index()
        return portfolio


def default_estimate_net_value(last_price: Decimal, analytic_portfolio: AnalyticPortfolio) -> float | None:
    with suppress(IndexError, ValueError):  # we silent any indexerror introduced by no returns for the past days
        return analytic_portfolio.get_estimate_net_value(float(last_price.net_value))


@receiver(post_save, sender="wbportfolio.PortfolioPortfolioThroughModel")
def post_portfolio_relationship_creation(sender, instance, created, raw, **kwargs):
    if (
        not raw
        and created
        and instance.portfolio.is_lookthrough
        and instance.type == PortfolioPortfolioThroughModel.Type.LOOK_THROUGH
    ):
        with suppress(AssetPosition.DoesNotExist):
            earliest_primary_position_date = instance.dependency_portfolio.assets.earliest("date").date
            compute_lookthrough_as_task.delay(instance.portfolio.id, earliest_primary_position_date, date.today())


@shared_task(queue=Queue.BACKGROUND.value)
def trigger_portfolio_change_as_task(portfolio_id, val_date, **kwargs):
    portfolio = Portfolio.all_objects.get(id=portfolio_id)
    portfolio.change_at_date(val_date, **kwargs)


@shared_task(queue=Queue.BACKGROUND.value)
def compute_lookthrough_as_task(portfolio_id: int, start: date, end: date):
    portfolio = Portfolio.objects.get(id=portfolio_id)
    portfolio.compute_lookthrough(start, to_date=end)


@receiver(investable_universe_updated, sender="wbfdm.Instrument")
def update_portfolio_after_investable_universe(*args, end_date: date | None = None, **kwargs):
    if not end_date:
        end_date = date.today()
    end_date = (end_date + timedelta(days=1) - BDay(1)).date()  # shift in case of business day
    from_date = (end_date - BDay(1)).date()
    excluded_positions = defaultdict(list)
    for portfolio in Portfolio.tracked_objects.all().to_dependency_iterator(from_date):
        if not portfolio.is_lookthrough:
            try:
                portfolio.propagate_or_update_assets(from_date, end_date)
                for positions in portfolio.builder.excluded_positions.values():
                    for pos in positions:
                        excluded_positions[pos.underlying_quote].append(portfolio)
                portfolio.builder.clear()
            except Exception as e:
                logger.error(
                    "Portfolio drift: Exception while handling portfolio.", extra={"portfolio": portfolio, "detail": e}
                )
        portfolio.estimate_net_asset_values(end_date)
    # if there were excluded positions, we compiled a itemized list of quote per portfolio that got excluded and warn the current portfolio manager
    if excluded_positions:
        body = (
            "<p>While drifting the portfolios, the following quotes got excluded because of missing prices: </p><ul>"
        )
        for quote, portfolios in excluded_positions.items():
            body += f"<li>{quote}</li><p>Impacted portfolios: </p><ul>"
            for portfolio in portfolios:
                body += f"<li>{portfolio}</li>"
            body += "</ul>"
        body += "</ul> <p>Note: If the quote has simply changed its primary exchange, please use the adjustment tool provided. Otherwise, please contact a system administrator.</p>"
        for user in User.objects.filter(profile__in=PortfolioRole.portfolio_managers(), is_active=True):
            send_notification(
                code="wbportfolio.portfolio.warning",
                title="Positions were automatically excluded",
                body=body,
                user=user,
            )



import numpy as np
import pandas as pd
from wbfdm.import_export.parsers.refinitiv.utils import _clean_and_return_dict

DEFAULT_MAPPING = {
    "AX": "factor",
    "Dates": "date",
    "Instrument": "instrument",
}


def parse(import_source):
    df = pd.read_json(import_source.file, orient="records")
    df = df.replace([np.inf, -np.inf, np.nan], None)
    df = df.rename(columns=DEFAULT_MAPPING).dropna(how="all", axis=1)
    df = df.dropna(how="all", subset=df.columns.difference(["instrument"]))
    df["date"] = pd.to_datetime(df["date"], utc=True, unit="ms")
    df.date = df.date.dt.strftime("%Y-%m-%d")
    df = df.dropna(how="any")
    data = list()
    if not df.empty:
        df = df.groupby(["instrument", "date"]).first().reset_index()
        data = _clean_and_return_dict(df)
    return {"data": data}



import rstr
from faker import Faker
from import_export import fields
from import_export.widgets import ForeignKeyWidget
from wbcore.contrib.io.resources import FilterModelResource
from wbfdm.models import Instrument

from wbportfolio.models import Trade

fake = Faker()


class OrderProposalTradeResource(FilterModelResource):
    """
    Trade Resource class to use to import trade from the order proposal
    """

    DUMMY_FIELD_MAP = {
        "underlying_instrument__isin": lambda: rstr.xeger("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})"),
        "underlying_instrument__ticker": "AAA",
        "underlying_instrument__name": "stock name",
        "weighting": 0.015,
        "shares": 1000.2536,
        "comment": lambda: fake.sentence(),
        "order": 1,
    }
    underlying_instrument__isin = fields.Field(
        column_name="underlying_instrument__isin",
        attribute="underlying_instrument",
        widget=ForeignKeyWidget(Instrument, field="isin"),
    )
    underlying_instrument__name = fields.Field(
        column_name="underlying_instrument__name",
        attribute="underlying_instrument",
        widget=ForeignKeyWidget(Instrument, field="name"),
    )
    underlying_instrument__ticker = fields.Field(
        column_name="underlying_instrument__ticker",
        attribute="underlying_instrument",
        widget=ForeignKeyWidget(Instrument, field="ticker"),
    )

    class Meta:
        fields = (
            "id",
            "underlying_instrument__isin",
            "underlying_instrument__name",
            "underlying_instrument__ticker",
            "weighting",
            "shares",
            "comment",
            "order",
        )
        export_order = fields
        model = Trade



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_manager

PORTFOLIOROLE_MENUITEM = MenuItem(
    label="Portfolio Roles",
    endpoint="wbportfolio:portfoliorole-list",
    permission=ItemPermission(permissions=["wbportfolio.view_portfoliorole"], method=is_manager),
    add=MenuItem(
        label="Create Portfolio Role",
        endpoint="wbportfolio:portfoliorole-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbportfolio.add_portfoliorole"]
        ),
    ),
)



from wbcore.metadata.configs.titles import TitleViewConfig


class FeesTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Fees"

    def get_instance_title(self):
        return "Fees: {{_product.name}} {{date}}"

    def get_create_title(self):
        return "New Fees"


class FeesProductTitleConfig(FeesTitleConfig):
    def get_list_title(self):
        return f"Fees: {self.view.product}"


class FeesAggregatedProductTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"Aggregated Fees for {self.view.product}"



# Generated by Django 5.0.14 on 2025-07-18 06:37

import django.db.models.deletion
import django.db.models.expressions
import django_fsm
from decimal import Decimal
from django.db import migrations, models
from tqdm import tqdm
def migrate_order(apps, schema_editor):
    Order = apps.get_model("wbportfolio", "Order")
    Trade = apps.get_model("wbportfolio", "Trade")
    objs = []
    qs = Trade.objects.filter(order_proposal__isnull=False).select_related("order_proposal")
    for order in tqdm(qs, total=qs.count()):
        objs.append(Order(
            order=order.order,
            value_date=order.order_proposal.trade_date,
            currency_fx_rate=order.currency_fx_rate,
            price=order.price,
            price_gross=order.price_gross,
            fees=order.fees,
            comment=order.comment,
            created=order.created,
            updated=order.updated,
            order_type=order.transaction_subtype,
            status=order.status,
            shares=order.shares,
            weighting=order.weighting,
            drift_factor=order.drift_factor,
            import_source=order.import_source,
            portfolio=order.portfolio,
            underlying_instrument=order.underlying_instrument,
            order_proposal=order.order_proposal,
        ))
    Order.objects.bulk_create(objs)
    qs.delete()

class Migration(migrations.Migration):

    dependencies = [
        ('io', '0008_importsource_resource_kwargs'),
        ('wbfdm', '0031_exchange_apply_round_lot_size_and_more'),
        ('wbportfolio', '0082_remove_tradeproposal_creator_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField(db_index=True, editable=False, verbose_name='order')),
                ('value_date', models.DateField(help_text='The date that this transaction was valuated/paid.', verbose_name='Value Date')),
                ('currency_fx_rate', models.DecimalField(decimal_places=8, default=Decimal('1'), max_digits=14, verbose_name='FOREX rate')),
                ('price', models.DecimalField(decimal_places=4, help_text='The price per share.', max_digits=16, verbose_name='Price')),
                ('price_gross', models.DecimalField(decimal_places=4, help_text='The gross price per share.', max_digits=16, verbose_name='Gross Price')),
                ('fees', models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(models.F('price_gross'), '-', models.F('price')), output_field=models.DecimalField(decimal_places=4, max_digits=20))),
                ('total_value_gross', models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(models.F('price_gross'), '*', models.F('shares')), output_field=models.DecimalField(decimal_places=4, max_digits=20))),
                ('total_value', models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(models.F('price'), '*', models.F('shares')), output_field=models.DecimalField(decimal_places=4, max_digits=20))),
                ('total_value_fx_portfolio', models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(django.db.models.expressions.CombinedExpression(models.F('currency_fx_rate'), '*', models.F('price')), '*', models.F('shares')), output_field=models.DecimalField(decimal_places=4, max_digits=20))),
                ('total_value_gross_fx_portfolio', models.GeneratedField(db_persist=True, expression=django.db.models.expressions.CombinedExpression(django.db.models.expressions.CombinedExpression(models.F('currency_fx_rate'), '*', models.F('price_gross')), '*', models.F('shares')), output_field=models.DecimalField(decimal_places=4, max_digits=20))),
                ('comment', models.TextField(blank=True, default='', verbose_name='Comment')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('order_type', models.CharField(choices=[('REBALANCE', 'Rebalance'), ('DECREASE', 'Decrease'), ('INCREASE', 'Increase'), ('BUY', 'Buy'), ('SELL', 'Sell'), ('NO_CHANGE', 'No Change')], default='BUY', max_length=32, verbose_name='Trade Type')),
                ('status', django_fsm.FSMField(choices=[('DRAFT', 'Draft'), ('SUBMIT', 'Submit'), ('EXECUTED', 'Executed'), ('CONFIRMED', 'Confirmed'), ('FAILED', 'Failed')], default='CONFIRMED', max_length=50, verbose_name='Status')),
                ('shares', models.DecimalField(decimal_places=4, default=Decimal('0.0'), help_text='The number of shares that were traded.', max_digits=15, verbose_name='Shares')),
                ('weighting', models.DecimalField(decimal_places=8, default=Decimal('0'), help_text='The weight to be multiplied against the target', max_digits=9, verbose_name='Weight')),
                ('drift_factor', models.DecimalField(decimal_places=16, default=Decimal('1'), help_text='Drift factor to be applied to the previous portfolio weight to get the actual effective weight including daily return', max_digits=19, verbose_name='Drift Factor')),
            ],
            options={
                'verbose_name': 'Order',
                'verbose_name_plural': 'Orders',
                'ordering': ('order',),
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='order',
            name='order_proposal',
            field=models.ForeignKey(help_text='The Order Proposal this trade is coming from',
                                    on_delete=django.db.models.deletion.CASCADE, related_name='orders',
                                    to='wbportfolio.orderproposal'),
        ),
        migrations.AddField(
            model_name='order',
            name='import_source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to='io.importsource'),
        ),
        migrations.AddField(
            model_name='order',
            name='portfolio',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='wbportfolio.portfolio', verbose_name='Portfolio'),
        ),
        migrations.AddField(
            model_name='order',
            name='underlying_instrument',
            field=models.ForeignKey(help_text='The instrument that is this transaction.',
                                    limit_choices_to=models.Q(('children__isnull', True)),
                                    on_delete=django.db.models.deletion.PROTECT, related_name='%(class)ss',
                                    to='wbfdm.instrument', verbose_name='Underlying Instrument'),
        ),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL IMMEDIATE;",
        ),
        migrations.RunPython(migrate_order),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL DEFERRED;",
        ),
        migrations.AlterModelOptions(
            name='trade',
            options={'verbose_name': 'Trade', 'verbose_name_plural': 'Trades'},
        ),
        migrations.RemoveConstraint(
            model_name='trade',
            name='unique_manual_trade',
        ),
        migrations.RemoveField(
            model_name='trade',
            name='drift_factor',
        ),
        migrations.RemoveField(
            model_name='trade',
            name='order',
        ),
        migrations.RemoveField(
            model_name='trade',
            name='order_proposal',
        ),
        migrations.AddIndex(
            model_name='order',
            index=models.Index(fields=['underlying_instrument', 'value_date'], name='wbportfolio_underly_051048_idx'),
        ),
        migrations.AddIndex(
            model_name='order',
            index=models.Index(fields=['portfolio', 'underlying_instrument', 'value_date'], name='wbportfolio_portfol_26d2a7_idx'),
        ),
        migrations.AddIndex(
            model_name='order',
            index=models.Index(fields=['order_proposal', 'underlying_instrument'], name='wbportfolio_order_p_637d28_idx'),
        ),
        migrations.AddConstraint(
            model_name='order',
            constraint=models.UniqueConstraint(fields=('order_proposal', 'underlying_instrument'), name='unique_order'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='comment',
            field=models.TextField(blank=True, default='', verbose_name='Order Comment'),
        ),
        migrations.RemoveField(
            model_name='order',
            name='drift_factor',
        ),
        migrations.AddField(
            model_name='order',
            name='daily_return',
            field=models.DecimalField(decimal_places=16, default=Decimal('0'), help_text='The Ex-Post daily return',
                                      max_digits=19, verbose_name='Daily Return'),
        ),
        migrations.AddIndex(
            model_name='order',
            index=models.Index(fields=['order_proposal'], name='wbportfolio_order_p_630213_idx'),
        ),
        migrations.RemoveField(
            model_name='order',
            name='status',
        ),
        migrations.RemoveField(
            model_name='trade',
            name='status',
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='status',
            field=django_fsm.FSMField(
                choices=[('DRAFT', 'Draft'), ('SUBMIT', 'Pending'), ('APPROVED', 'Approved'), ('DENIED', 'Denied'),
                         ('FAILED', 'Failed')], default='DRAFT', max_length=50, verbose_name='Status'),
        ),
    ]



from wbcore.metadata.configs.titles import TitleViewConfig

from wbportfolio.models import Product


class PortfolioRoleTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Portfolio Roles"

    def get_instance_title(self):
        return "Portfolio Role"

    def get_create_title(self):
        return "Create New Role"


class PortfolioRoleInstrumentTitleConfig(PortfolioRoleTitleConfig):
    def get_list_title(self):
        product = Product.objects.get(id=self.view.kwargs["instrument_id"])
        return f"Portfolio Role for {product.name}"

    def get_create_title(self):
        product = Product.objects.get(id=self.view.kwargs["instrument_id"])
        return f"New Portfolio Role for {product.name}"



from wbcore import viewsets
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin

from wbportfolio.models import Register
from wbportfolio.serializers import (
    RegisterModelSerializer,
    RegisterRepresentationSerializer,
)
from wbportfolio.viewsets.configs.buttons.registers import RegisterButtonConfig
from wbportfolio.viewsets.configs.display.registers import RegisterDisplayConfig
from wbportfolio.viewsets.configs.titles.registers import RegisterTitleConfig


class RegisterRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    serializer_class = RegisterRepresentationSerializer

    queryset = Register.objects.all()
    ordering = ["register_reference"]
    search_fields = [
        "register_reference",
        "register_name_1",
        "register_name_2",
        "custodian_reference",
        "custodian_name_1",
        "custodian_name_2",
        "outlet_reference",
        "outlet_name",
    ]


class RegisterModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    queryset = Register.objects.all()
    serializer_class = RegisterModelSerializer

    filterset_fields = {
        "register_reference": ["icontains"],
        "register_name_1": ["icontains"],
        "register_name_2": ["icontains"],
        "custodian_reference": ["icontains"],
        "custodian_name_1": ["icontains"],
        "custodian_name_2": ["icontains"],
        "outlet_reference": ["icontains"],
        "outlet_name": ["icontains"],
        "status": ["exact"],
        "status_message": ["icontains"],
        "global_register_reference": ["exact"],
    }
    ordering = ["register_reference"]
    search_fields = [
        "register_reference",
        "register_name_1",
        "register_name_2",
        "custodian_reference",
        "custodian_name_1",
        "custodian_name_2",
        "outlet_reference",
        "outlet_name",
    ]

    display_config_class = RegisterDisplayConfig
    title_config_class = RegisterTitleConfig
    button_config_class = RegisterButtonConfig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related(
                "custodian_country",
                "custodian_city",
                "outlet_city",
                "outlet_country",
                "citizenship",
                "residence",
            )
        )



from datetime import date, timedelta

from rest_framework.exceptions import ValidationError
from wbcore import serializers

from wbportfolio.models.portfolio_cash_flow import DailyPortfolioCashFlow
from wbportfolio.serializers.portfolios import PortfolioRepresentationSerializer


class DailyPortfolioCashFlowModelSerializer(serializers.ModelSerializer):
    swing_pricing_indicator = serializers.CharField(read_only=True)
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")

    def validate(self, data):
        # We explictly forbit changing any data that is older than 1 week
        if value_date := data.get("value_date", getattr(self.instance, "value_date", None)):
            if value_date < (date.today() - timedelta(days=7)) and not self.context["request"].user.has_perm(
                "administrate_dailyportfoliocashflow"
            ):
                raise ValidationError({"non_field_errors": ["Changing entries older than 1 week is prohibited."]})

        return data

    class Meta:
        model = DailyPortfolioCashFlow
        percent_fields = ["true_cash_pct", "cash_pct", "target_cash_pct", "cash_flow_asset_ratio"]
        override_decimal_places = 0
        fields = (
            "id",
            "value_date",
            "portfolio",
            "_portfolio",
            "cash",
            "cash_pct",
            "cash_flow_forecast",
            "total_assets",
            "estimated_total_assets",
            "cash_flow_asset_ratio",
            "swing_pricing_indicator",
            "true_cash",
            "true_cash_pct",
            "target_cash",
            "target_cash_pct",
            "excess_cash",
            "proposed_rebalancing",
            "rebalancing",
            "pending",
            "comment",
        )



from wbcore.menus import ItemPermission, MenuItem

CONSOLIDATED_TRADE_SUMMARY_TABLE_MENUITEM = MenuItem(
    label="Consolidated Trade Summary",
    endpoint="wbportfolio:aumtable-list",
    permission=ItemPermission(permissions=["wbportfolio.view_claim"], method=lambda request: request.user.is_internal),
)

PNL_MENUITEM = MenuItem(
    label="Profit & Loss",
    endpoint="wbportfolio:pnltable-list",
    permission=ItemPermission(permissions=["wbportfolio.view_claim"], method=lambda request: request.user.is_internal),
)

CLAIM_MENUITEM = MenuItem(
    label="Claims",
    endpoint="wbportfolio:claim-list",
    permission=ItemPermission(permissions=["wbportfolio.view_claim"]),
    add=MenuItem(
        label="Create Claim",
        endpoint="wbportfolio:claim-list",
        permission=ItemPermission(permissions=["wbportfolio.add_claim"]),
    ),
)

NEGATIVEACCOUNTPRODUCT_MENUITEM = MenuItem(
    label="Negative Account per Product",
    endpoint="wbportfolio:negativeaccountproduct-list",
    permission=ItemPermission(method=lambda request: request.user.is_internal),
)

AUM_NNM_PROGRESSION_MENUITEM = MenuItem(
    label="AUM/NNM Progression",
    endpoint="wbportfolio:assetandnetnewmoneyprogression-list",
    permission=ItemPermission(method=lambda request: request.user.is_internal),
)



from django.db import models
from django.db.models import Q
from wbcore.models import WBModel


class PortfolioSwingPricing(WBModel):
    valid_date = models.DateField()

    portfolio = models.ForeignKey(
        to="wbportfolio.Portfolio",
        related_name="swing_pricings",
        on_delete=models.CASCADE,
    )

    negative_threshold = models.DecimalField(max_digits=4, decimal_places=4)
    negative_swing_factor = models.DecimalField(max_digits=4, decimal_places=4)
    positive_threshold = models.DecimalField(max_digits=4, decimal_places=4)
    positive_swing_factor = models.DecimalField(max_digits=4, decimal_places=4)

    def __str__(self) -> str:
        return f"{self.portfolio}: {self.negative_swing_factor:.2%}/{self.positive_swing_factor:.2%} ({self.valid_date:%d.%m.%Y})"

    class Meta:
        verbose_name = "Portfolio Swing Pricing"
        verbose_name_plural = "Portfolio Swing Pricings"
        constraints = [
            models.CheckConstraint(
                condition=Q(negative_threshold__lt=0)
                & Q(negative_swing_factor__lt=0)
                & Q(positive_threshold__gt=0)
                & Q(positive_swing_factor__gt=0),
                name="value_polarity",
            ),
            models.UniqueConstraint(fields=["valid_date", "portfolio"], name="unique_valid_date_portfolio"),
        ]

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfolioswingpricing"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:portfolioswingpricing-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{portfolio}}: {{swing_factor}} ({{valid_date}})"



import factory
from faker import Faker
from pandas._libs.tslibs.offsets import BDay
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.currency.factories import CurrencyFactory

from wbportfolio.factories import PortfolioFactory
from wbportfolio.models import OrderProposal

fake = Faker()


class OrderProposalFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = OrderProposal

    trade_date = factory.LazyAttribute(lambda o: (fake.date_object() + BDay(1)).date())
    comment = factory.Faker("paragraph")
    portfolio = factory.LazyAttribute(lambda o: PortfolioFactory.create(currency=CurrencyFactory.create(key="USD")))
    creator = factory.LazyAttribute(lambda o: UserFactory.create().profile)
    approver = factory.LazyAttribute(lambda o: UserFactory.create().profile)



import networkx as nx
import plotly.graph_objects as go
from networkx.drawing.nx_agraph import graphviz_layout


def reformat_graph_layout(g, layout):
    """
    this method provide positions based on layout algorithm
    :param g:
    :param layout:
    :return:
    """
    if layout == "graphviz":
        positions = graphviz_layout(g)
    elif layout == "spring":
        positions = nx.fruchterman_reingold_layout(g, k=0.5, iterations=1000)
    elif layout == "spectral":
        positions = nx.spectral_layout(g, scale=0.1)
    elif layout == "random":
        positions = nx.random_layout(g)
    else:
        raise Exception("please specify the layout from graphviz, spring, spectral or random")

    return positions


def networkx_graph_to_plotly(
    g: nx.Graph,
    labels: dict[str, str] | None = None,
    node_size: int = 10,
    edge_weight: int = 1,
    edge_color: str = "black",
    layout: str = "graphviz",
    title: str = "",
) -> go.Figure:
    """
    Visualize a NetworkX graph using Plotly.
    """
    positions = reformat_graph_layout(g, layout)
    if not labels:
        labels = {}
    # Initialize edge traces
    edge_traces = []
    for edge in g.edges():
        x0, y0 = positions[edge[0]]
        x1, y1 = positions[edge[1]]

        edge_trace = go.Scatter(
            x=[x0, x1], y=[y0, y1], line=dict(width=edge_weight, color=edge_color), hoverinfo="none", mode="lines"
        )
        edge_traces.append(edge_trace)

    # Initialize node trace
    node_x, node_y, node_colors, node_labels = [], [], [], []
    for node in g.nodes():
        x, y = positions[node]
        node_x.append(x)
        node_y.append(y)
        node_labels.append(labels.get(node, node))
        node_colors.append(len(list(g.neighbors(node))))  # Color based on degree

    node_trace = go.Scatter(
        x=node_x,
        y=node_y,
        text=node_labels,
        mode="markers+text",
        textfont=dict(family="Calibri (Body)", size=15, color="grey"),
        marker=dict(
            size=node_size,
            color=node_colors,
        ),
    )

    # Assemble the figure
    fig = go.Figure(
        data=edge_traces + [node_trace],
        layout=go.Layout(
            showlegend=False,
            template="plotly_white",
            margin=dict(l=50, r=50, t=0, b=40),
        ),
    )
    return fig



import numpy as np
import pandas as pd

from .utils import get_perf_fee_isin

FIELD_MAP = {
    "EndOfDay": "date",
    "LastPriceDate": "asset_valuation_date",
    "LastPrice": "initial_price",
    "Quantity": "initial_shares",
    "TradingCurrency": "currency__key",
    "FxRate": "initial_currency_fx_rate",
    "ISIN": "underlying_quote__isin",
    "InstrumentName": "underlying_quote__name",
    "BuCurrency": "portfolio__currency__key",
    "Ticker": "underlying_quote__ticker",
    # "TradingPlace": "exchange",
    "Booking_comment": "comment",
    "AssetGroup": "underlying_quote__instrument_type",
    "InstrumentIdentifier": "underlying_quote__identifier",
    "#ClientName": "portfolio__identifier",
}

ASSET_GROUP_MAP = {
    "bond": "bond",
    "cash": "cash",
    "comdty": "commodity",
    "credit": "credit",
    "fund": "exchange_trade_product",
    "fut": "future",
    "fwd": "forward",
    "idx": "index",
    "insurpol": "insurpol",
    "interrateopt": "interest_rate_derivative",
    "limit": "limit",
    "opt": "option",
    "struct": "product",
    "swap": "swap",
    "cash_equivalent": "cash_equivalent",
}


def parse(import_source):
    perf_fee_isin = get_perf_fee_isin(import_source.source)
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=";")
    df = df.rename(columns=FIELD_MAP).replace([np.inf, -np.inf, np.nan], None)
    df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
    if import_source.source.import_parameters.get("import_perf_fees_as_cash", False):
        fees_idx = df["underlying_quote__isin"] == perf_fee_isin
        df.loc[fees_idx, "underlying_quote__isin"] = None
        df.loc[fees_idx, "underlying_quote__instrument_type"] = "cash_equivalent"
        df.loc[fees_idx, "initial_shares"] = df.loc[fees_idx, "initial_price"]  # inverse price and share
    else:
        df = df.loc[df["underlying_quote__isin"] != perf_fee_isin, :]

    df["underlying_quote__instrument_type"] = df["underlying_quote__instrument_type"].apply(
        lambda x: ASSET_GROUP_MAP[x] if x else None
    )
    # we remove instrument without isin or not of type cash because we don't support these (e.g, Dividend)
    df = df[
        ~df["underlying_quote__isin"].isnull()
        | df["underlying_quote__instrument_type"].isin(["cash", "cash_equivalent"])
    ]

    # we group by instrument identifier to be sure we don't introduce duplicated position
    df = (
        df.groupby("underlying_quote__identifier", dropna=False)
        .agg(
            {
                "initial_currency_fx_rate": "first",
                "initial_price": "first",
                "initial_shares": "sum",
                "date": "first",
                "asset_valuation_date": "first",
                "currency__key": "first",
                "underlying_quote__isin": "first",
                "underlying_quote__name": "first",
                "portfolio__currency__key": "first",
                "underlying_quote__ticker": "first",
                "underlying_quote__instrument_type": "first",
                "portfolio__identifier": "first",
            }
        )
        .reset_index(drop=True)
    )
    df.loc[df["underlying_quote__instrument_type"].isin(["cash", "cash_equivalent"]), "initial_price"] = 1.0
    df["date"] = pd.to_datetime(df["date"], dayfirst=True).dt.strftime("%Y-%m-%d")
    df["asset_valuation_date"] = pd.to_datetime(df["asset_valuation_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
    df["is_estimated"] = False
    df["underlying_quote__currency__key"] = df["currency__key"]
    df["weighting"] = df.initial_currency_fx_rate * df.initial_price * df.initial_shares
    df["weighting"] = df.weighting / df.weighting.sum()
    df["portfolio__currency__key"] = (
        df["portfolio__currency__key"].ffill().bfill()
    )  # We do that because we expect all cell value to be the same but we expect also emtpy cell (mistake from the provider)
    df = df.replace([np.inf, -np.inf, np.nan], None)
    return {"data": df.to_dict("records")}



from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class AdjustmentButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(key="detection_table", label="Detection Table", icon=WBIcon.DATA_GRID.icon),
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



from datetime import date

from celery import shared_task
from django.db.models import Exists, OuterRef
from tqdm import tqdm
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Company
from wbcore.workers import Queue
from wbcrm.models import Account

from .models import CompanyPortfolioData, Updater


@shared_task(queue=Queue.BACKGROUND.value)
def update_all_portfolio_data(val_date: date | None = None):
    if not val_date:
        val_date = CurrencyFXRates.objects.latest("date").date
    updater = Updater(val_date)
    qs = Company.objects.annotate(
        has_account=Exists(Account.objects.filter(owner=OuterRef("pk"))),
        has_portfolio_data=Exists(CompanyPortfolioData.objects.filter(company=OuterRef("pk"))),
    )
    company_objs = []
    portfolio_data_objs = []
    for company in tqdm(qs, total=qs.count()):
        portfolio_data = CompanyPortfolioData.objects.get_or_create(company=company)[0]
        company.customer_status, company.tier = updater.update_company_data(portfolio_data)
        portfolio_data_objs.append(portfolio_data)
        company_objs.append(company)
    if company_objs:
        Company.objects.bulk_update(company_objs, ["customer_status", "tier"])
    if portfolio_data_objs:
        CompanyPortfolioData.objects.bulk_update(
            portfolio_data_objs, ["invested_assets_under_management_usd", "potential"]
        )



# Generated by Django 5.2.9 on 2026-01-21 12:27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0037_relatedinstrumentthroughmodel_unique_instrument_relationship_and_more'),
        ('wbportfolio', '0095_portfolio_keep_only_essential_positions'),
    ]

    operations = [
        migrations.CreateModel(
            name='PortfolioTradableExchangeThroughModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('rebalancing_adjustment_factor', models.FloatField(default=0)),
                ('exposure_adjustment_factor', models.FloatField(default=0)),
                ('exchange', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tradable_portfolio_relationships', to='wbfdm.exchange')),
                ('portfolio', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tradable_exchange_relationships', to='wbportfolio.portfolio')),
            ],
            options={
                'verbose_name': 'Portfolio Tradeable Exchange',
                'verbose_name_plural': 'Portfolio Tradeable Exchanges',
                'constraints': [models.UniqueConstraint(fields=('portfolio', 'exchange'), name='unique_tradable_exchange_per_portfolio')],
            },
        ),
    ]



from django.conf import settings

from wbportfolio.order_routing import ExecutionStatus
from wbportfolio.order_routing.adapters import BaseCustodianAdapter
from wbportfolio.pms.typing import Order


class Router:
    def __init__(self, adapter: BaseCustodianAdapter):
        self.adapter = adapter

    @property
    def submit_as_draft(self):
        return getattr(settings, "DEBUG", True) or getattr(settings, "ORDER_ROUTING_AS_DRAFT", True)

    def submit_rebalancing(self, orders: list[Order]) -> tuple[list[Order], tuple[str, str]]:
        """
        Submit a rebalance order for the certificate.
        """
        items = self.adapter.serialize_orders(orders)
        confirmed_items, msg = self.adapter.submit_rebalancing(items, as_draft=self.submit_as_draft)
        status = ExecutionStatus.IN_DRAFT if self.submit_as_draft else ExecutionStatus.PENDING
        return self.adapter.deserialize_items(confirmed_items), (status, msg)

    def get_rebalance_status(self) -> tuple[ExecutionStatus, str]:
        return self.adapter.get_rebalance_status()

    def cancel_rebalancing(self) -> bool:
        return self.adapter.cancel_current_rebalancing()

    def get_current_rebalancing_request(self) -> list[Order]:
        items = self.adapter.get_current_rebalancing()
        return self.adapter.deserialize_items(items)



from django.contrib import admin

from wbportfolio.models import InstrumentPortfolioThroughModel
from wbportfolio.models.portfolio_relationship import (
    PortfolioInstrumentPreferredClassificationThroughModel,
    PortfolioTradableExchangeThroughModel,
)


class PortfolioInstrumentPreferredClassificationThroughInlineModelAdmin(admin.TabularInline):
    model = PortfolioInstrumentPreferredClassificationThroughModel
    fk_name = "portfolio"
    raw_id_fields = [
        "portfolio",
        "instrument",
        "classification",
        "classification_group",
    ]


class InstrumentPortfolioThroughModelAdmin(admin.TabularInline):
    model = InstrumentPortfolioThroughModel
    fk_name = "instrument"
    autocomplete_fields = ["portfolio"]


class PortfolioTradableExchangeThroughModelInLineModelAdmin(admin.TabularInline):
    model = PortfolioTradableExchangeThroughModel
    autocomplete_fields = ["exchange"]
    fk_name = "portfolio"
    raw_id_fields = [
        "portfolio",
        "exchange",
    ]



from typing import Generator

import pandas as pd
from wbcompliance.models import RiskIncidentType
from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcore import serializers as wb_serializers
from wbcore.serializers.fields.number import percent_decorator
from wbfdm.models import Instrument

from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from .mixins import ActivePortfolioRelationshipMixin


@register("UCITS 5|10|40 Portfolio Rule Backend", rule_group_key="portfolio")
class RuleBackend(ActivePortfolioRelationshipMixin, backend.AbstractRuleBackend):
    DEFAULT_THRESHOLD_1: float = 0.05
    DEFAULT_THRESHOLD_2: float = 0.10
    DEFAULT_THRESHOLD_3: float = 0.40

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            threshold_1 = wb_serializers.FloatField(
                default=cls.DEFAULT_THRESHOLD_1, label="Threshold 1", percent=True, decorators=[percent_decorator]
            )
            threshold_2 = wb_serializers.FloatField(
                default=cls.DEFAULT_THRESHOLD_2, label="Threshold 2", percent=True, decorators=[percent_decorator]
            )
            threshold_3 = wb_serializers.FloatField(
                default=cls.DEFAULT_THRESHOLD_3, label="Threshold 3", percent=True, decorators=[percent_decorator]
            )

            @classmethod
            def get_parameter_fields(cls):
                return ["threshold_1", "threshold_2", "threshold_3"]

        return RuleBackendSerializer

    def _filter_df(self, df):
        if df.empty:
            return df
        return df[(df["weighting"] >= self.threshold_1) & (~df["is_cash"])]

    def _process_dto(self, portfolio: PortfolioDTO, **kwargs) -> Generator[backend.IncidentResult, None, None]:
        if not (df := self._filter_df(pd.DataFrame(portfolio.to_df()).astype({"weighting": float}))).empty:
            df = df[["underlying_instrument", "weighting"]].groupby("underlying_instrument").sum()
            total_weight_threshold_1_2 = df["weighting"].sum()
            highest_incident_type = RiskIncidentType.objects.order_by("-severity_order").first()

            for id, row in df.to_dict("index").items():
                if (row["weighting"] > self.threshold_2) or (total_weight_threshold_1_2 > self.threshold_3):
                    instrument = Instrument.objects.get(id=id)
                    breached_value = f"""
                    Sum >= {self.threshold_1:.2%}: {total_weight_threshold_1_2:+.2%}
                    """
                    if row["weighting"] > self.threshold_2:
                        breached_value += f"<br>Instrument >= {self.threshold_2:.2%}: {instrument.name_repr}"
                    yield backend.IncidentResult(
                        breached_object=instrument,
                        breached_object_repr=str(instrument),
                        breached_value=breached_value,
                        report_details={
                            "Breach Thresholds": f"{self.threshold_1:.2%}|{self.threshold_2:.2%}|{self.threshold_3:.2%}",
                            "Weighting": f"{row['weighting']:+.2%}",
                            f"Sum of positions > {self.threshold_1:.2%}": f"{total_weight_threshold_1_2:+.2%}",
                        },
                        severity=highest_incident_type,
                    )



from wbcore.menus import ItemPermission, MenuItem

REGISTER_MENUITEM = MenuItem(
    label="Registers",
    endpoint="wbportfolio:register-list",
    permission=ItemPermission(
        permissions=["wbportfolio.view_register"], method=lambda request: request.user.is_internal
    ),
)



from wbcore.menus import ItemPermission, MenuItem

CustodianMenuItem = MenuItem(
    label="Custodian",
    endpoint="wbportfolio:custodian-list",
    permission=ItemPermission(
        permissions=["wbportfolio.view_custodian"],
        method=lambda request: request.user.is_internal,
    ),
)



from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)

from wbportfolio.models import Custodian
from wbportfolio.serializers import CustodianRepresentationSerializer


class CustodianButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        custodians_qs = Custodian.objects.all()
        if self.view.kwargs.get("pk", None):
            custodians_qs = custodians_qs.exclude(id=self.view.kwargs["pk"])

        class CustodianMergeModelSerializer(wb_serializers.ModelSerializer):
            merged_custodian = wb_serializers.PrimaryKeyRelatedField(
                required=True, queryset=custodians_qs, label="Custodian to be merged"
            )
            _merged_custodian = CustodianRepresentationSerializer(source="merged_custodian")

            class Meta:
                model = Custodian
                fields = ("id", "merged_custodian", "_merged_custodian")

        return {
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:custodian",),
                key="merge",
                icon=WBIcon.TRADE.icon,
                label="Merge custodians",
                description_fields="""
                <p>Merge the custodian <strong>{{name}}</strong> with another custodian in the list bellow</p>
                """,
                serializer=CustodianMergeModelSerializer,
                action_label="Merge",
                title="Merge custodians",
                instance_display=create_simple_display([["merged_custodian"]]),
            )
        }

    def get_custom_instance_buttons(self):
        if custodian_id := self.view.kwargs.get("pk", None):
            custodian = Custodian.objects.get(id=custodian_id)

            class AdjustmentModelSerializer(wb_serializers.ModelSerializer):
                mapping_choice = wb_serializers.ChoiceField(
                    required=True, choices={m: m for m in custodian.mapping}, label="Mapping to split off"
                )

                class Meta:
                    model = Custodian
                    fields = ("id", "mapping_choice", "name")

            merge_button = {
                bt.ActionButton(
                    method=RequestType.PATCH,
                    identifiers=("wbportfolio:custodian",),
                    key="split",
                    icon=WBIcon.TRADE.icon,
                    label="Adjustment custodians",
                    description_fields="""
                    <p>Adjustment the custodian <strong>{{name}}</strong> with another custodian in the list bellow</p>
                    """,
                    serializer=AdjustmentModelSerializer,
                    action_label="Adjustment",
                    title="Adjustment custodians",
                    instance_display=create_simple_display([["mapping_choice"]]),
                )
            }
            return self.get_custom_list_instance_buttons() | merge_button
        return self.get_custom_list_instance_buttons()



# Generated by Django 5.0.10 on 2024-12-16 15:32

from django.db import migrations, models
import django.db.models.deletion
from datetime import date

from django.db.models import Exists


def migrate_path(apps, schema_editor):
    Product = apps.get_model("wbportfolio", "Product")
    Index = apps.get_model("wbportfolio", "Index")
    for product in Product.objects.filter(price_computation__import_path="restbench.price.peers"):
        product.custom_net_asset_value_computation_method_path = "restbench.price.peers" + ".get_estimate_net_value"
        product.save()
    for index in Index.objects.filter(price_computation__import_path="restbench.price.peers"):
        index.custom_net_asset_value_computation_method_path = "restbench.price.peers" + ".get_estimate_net_value"
        index.save()


def migrate_data(apps, schema_editor):
    Portfolio = apps.get_model("wbportfolio", "Portfolio")
    RebalancingModel = apps.get_model("wbportfolio", "RebalancingModel")
    PortfolioPortfolioThroughModel = apps.get_model("wbportfolio", "PortfolioPortfolioThroughModel")
    Rebalancer = apps.get_model("wbportfolio", "Rebalancer")
    equally_weighted_model, _ = RebalancingModel.objects.get_or_create(
        class_path="wbportfolio.rebalancing.models.equally_weighted.EquallyWeightedRebalancing",
        defaults={"name": "Equally Weighted Rebalancing"},
    )
    composite_model, _ = RebalancingModel.objects.get_or_create(
        class_path="wbportfolio.rebalancing.models.composite.CompositeRebalancing",
        defaults={"name": "Composite Rebalancing"},
    )
    PortfolioPortfolioThroughModel.objects.filter(portfolio=models.F("dependency_portfolio")).delete()
    Portfolio.objects.annotate(has_primary=Exists(PortfolioPortfolioThroughModel.objects.filter(portfolio=models.OuterRef("pk"), type="PRIMARY"))).filter(has_primary=True, portfolio_synchronization__name="Default Portfolio Synchronization").update(
        is_lookthrough=True
    )
    for portfolio in Portfolio.objects.filter(portfolio_synchronization__name="Composite synchronization"):
        try:
            activation_date = portfolio.assets.earliest("date").date
        except Exception:
            activation_date = date.today()
        Rebalancer.objects.create(
            portfolio=portfolio,
            rebalancing_model=composite_model,
            approve_trade_proposal_automatically=True,
            frequency="RRULE:FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1",
            activation_date=activation_date,
        )
    for portfolio in Portfolio.objects.filter(portfolio_synchronization__name="Equally-weighted Synchronization"):
        try:
            activation_date = portfolio.assets.earliest("date").date
        except Exception:
            activation_date = date.today()
        Rebalancer.objects.create(
            portfolio=portfolio,
            rebalancing_model=equally_weighted_model,
            approve_trade_proposal_automatically=True,
            frequency="RRULE:FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1",
            activation_date=activation_date,
        )

def set_composition(apps, schema_editor):
    Portfolio = apps.get_model("wbportfolio", "Portfolio")
    for portfolio in Portfolio.objects.all():
        if portfolio.assets.filter(underlying_instrument__instrument_type__key__in=["index", "product", "product_group"]).exists():
            portfolio.is_composition = True
            portfolio.save()

class Migration(migrations.Migration):
    dependencies = [
        ('directory', '0009_remove_entry_external_identfier_and_more'),
        ('io', '0007_alter_exportsource_query_params'),
        ('wbfdm', '0028_instrumentprice_annualized_daily_volatility'),
        ("wbportfolio", "0072_trade_diff_shares"),
    ]

    operations = [
        migrations.AddField(
            model_name="index",
            name="custom_net_asset_value_computation_method_path",
            field=models.CharField(blank=True, null=True, verbose_name="Custom NAV Computation Method"),
        ),
        migrations.AddField(
            model_name="product",
            name="custom_net_asset_value_computation_method_path",
            field=models.CharField(blank=True, null=True, verbose_name="Custom NAV Computation Method"),
        ),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL IMMEDIATE;",
        ),
        migrations.RunPython(migrate_path),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL DEFERRED;",
        ),
        migrations.RemoveField(
            model_name="product",
            name="price_computation",
        ),
        migrations.RemoveField(
            model_name="index",
            name="price_computation",
        ),
        migrations.DeleteModel(
            name="PriceComputation",
        ),
        migrations.CreateModel(
            name="RebalancingModel",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=64, verbose_name="Name")),
                ("class_path", models.CharField(max_length=512, verbose_name="Class path")),
            ],
            options={
                "verbose_name": "Rebalancing Model",
                "verbose_name_plural": "Rebalancing Models",
            },
        ),
        migrations.RenameField(
            model_name="portfolio",
            old_name="last_synchronization",
            new_name="updated_at",
        ),
        migrations.AddField(
            model_name="portfolio",
            name="is_lookthrough",
            field=models.BooleanField(
                default=False, help_text="Indicates that this portfolio is a look-through portfolio"
            ),
        ),
        migrations.AddField(
            model_name="tradeproposal",
            name="rebalancing_model",
            field=models.ForeignKey(
                blank=True,
                help_text="Rebalancing Model to use to generate the target portfolio",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="trade_proposals",
                to="wbportfolio.rebalancingmodel",
                verbose_name="Rebalancing Model",
            ),
        ),
        migrations.CreateModel(
            name="Rebalancer",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("parameters", models.JSONField(default=dict, verbose_name="Parameters")),
                (
                    "approve_trade_proposal_automatically",
                    models.BooleanField(default=False, verbose_name="Apply Trade Proposal Automatically"),
                ),
                ("activation_date", models.DateField(verbose_name="Activation Date")),
                (
                    "frequency",
                    models.CharField(
                        default="RRULE:FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1",
                        help_text="The Evaluation Frequency in RRULE format",
                        max_length=56,
                        verbose_name="Evaluation Frequency",
                    ),
                ),
                (
                    "portfolio",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="automatic_rebalancer",
                        to="wbportfolio.portfolio",
                    ),
                ),
                (
                    "rebalancing_model",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="rebalancers",
                        to="wbportfolio.rebalancingmodel",
                        verbose_name="Rebalancing Model",
                    ),
                ),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
            ],
            options={
                "verbose_name": "Rebalancer",
                "verbose_name_plural": "Rebalancers",
            },
        ),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL IMMEDIATE;",
        ),
        migrations.RunPython(migrate_data),
        migrations.RunSQL(
            sql="SET CONSTRAINTS ALL DEFERRED;",
        ),
        migrations.RemoveField(
            model_name="tradeproposal",
            name="model_portfolio",
        ),
        migrations.RemoveField(
            model_name="portfolio",
            name="portfolio_synchronization",
        ),
        migrations.RemoveField(
            model_name="synchronizationtask",
            name="dependent_task",
        ),
        migrations.RemoveField(
            model_name="synchronizationtask",
            name="periodictask_ptr",
        ),
        migrations.DeleteModel(
            name="PortfolioSynchronization",
        ),
        migrations.DeleteModel(
            name="SynchronizationTask",
        ),
        migrations.AlterUniqueTogether(
            name='adjustment',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='instrumentportfoliothroughmodel',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='portfolioinstrumentpreferredclassificationthroughmodel',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='tradeproposal',
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name='index',
            name='custom_net_asset_value_computation_method_path',
        ),
        migrations.RemoveField(
            model_name='product',
            name='custom_net_asset_value_computation_method_path',
        ),
        migrations.AddField(
            model_name='index',
            name='net_asset_value_computation_method_path',
            field=models.CharField(blank=True, default='wbportfolio.models.portfolio.default_estimate_net_value',
                                   null=True, verbose_name='NAV Computation Method'),
        ),
        migrations.AddField(
            model_name='product',
            name='net_asset_value_computation_method_path',
            field=models.CharField(blank=True, default='wbportfolio.models.portfolio.default_estimate_net_value',
                                   null=True, verbose_name='NAV Computation Method'),
        ),
        migrations.AlterField(
            model_name='portfolio',
            name='is_tracked',
            field=models.BooleanField(default=True,
                                      help_text='True if the internal updating mechanism (e.g., Next weights or Look-Through computation, rebalancing etc...) needs to apply to this portfolio'),
        ),
        migrations.AlterField(
            model_name='portfolio',
            name='updated_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Updated At'),
        ),
        migrations.AlterField(
            model_name='tradeproposal',
            name='rebalancing_model',
            field=models.ForeignKey(blank=True, help_text='Rebalancing Model that generates the target portfolio',
                                    null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='trade_proposals', to='wbportfolio.rebalancingmodel',
                                    verbose_name='Rebalancing Model'),
        ),
        migrations.AddConstraint(
            model_name='adjustment',
            constraint=models.UniqueConstraint(fields=('date', 'instrument'), name='unique_date_instrument'),
        ),
        migrations.AddConstraint(
            model_name='portfolioinstrumentpreferredclassificationthroughmodel',
            constraint=models.UniqueConstraint(fields=('portfolio', 'instrument', 'classification_group'),
                                               name='unique_prefered_classification_relationship'),
        ),
        migrations.AddConstraint(
            model_name='tradeproposal',
            constraint=models.UniqueConstraint(fields=('portfolio', 'trade_date'), name='unique_trade_proposal'),
        ),
        migrations.AddField(
            model_name='productgroup',
            name='net_asset_value_computation_method_path',
            field=models.CharField(blank=True, default='wbportfolio.models.portfolio.default_estimate_net_value',
                                   null=True, verbose_name='NAV Computation Method'),
        ),
        migrations.AddField(
            model_name='productgroup',
            name='risk_scale',
            field=models.IntegerField(default=4, validators=[django.core.validators.MinValueValidator(1),
                                                             django.core.validators.MaxValueValidator(7)],
                                      verbose_name='Risk Scale'),
        ),
        migrations.AlterField(
            model_name='index',
            name='risk_scale',
            field=models.IntegerField(default=4, validators=[django.core.validators.MinValueValidator(1),
                                                             django.core.validators.MaxValueValidator(7)],
                                      verbose_name='Risk Scale'),
        ),
        migrations.AlterField(
            model_name='index',
            name='net_asset_value_computation_method_path',
            field=models.CharField(blank=True, default='wbportfolio.models.portfolio.default_estimate_net_value',
                                   null=True, verbose_name='NAV Computation Method'),
        ),
        migrations.AlterField(
            model_name='product',
            name='net_asset_value_computation_method_path',
            field=models.CharField(blank=True, default='wbportfolio.models.portfolio.default_estimate_net_value',
                                   null=True, verbose_name='NAV Computation Method'),
        ),
        migrations.AlterField(
            model_name='productgroup',
            name='net_asset_value_computation_method_path',
            field=models.CharField(blank=True, default='wbportfolio.models.portfolio.default_estimate_net_value',
                                   null=True, verbose_name='NAV Computation Method'),
        ),
        migrations.RemoveConstraint(
            model_name='assetposition',
            name='wbportfolio_assetposition_weekday_constraint',
        ),
        migrations.RemoveConstraint(
            model_name='assetposition',
            name='unique_asset_position',
        ),
        migrations.RemoveIndex(
            model_name='assetposition',
            name='wbportfolio_date_0a7c0a_idx',
        ),
        migrations.RemoveIndex(
            model_name='assetposition',
            name='wbportfolio_date_da82d2_idx',
        ),
        migrations.AddField(
            model_name='assetposition',
            name='underlying_quote',
            field=models.ForeignKey(blank=True, help_text='The quote that is this asset.',
                                    limit_choices_to=models.Q(('children__isnull', True)), null=True,
                                    on_delete=django.db.models.deletion.PROTECT, related_name='assets',
                                    to='wbfdm.instrument', verbose_name='Underlying Quote'),
        ),
        migrations.AlterField(
            model_name='assetposition',
            name='underlying_instrument',
            field=models.ForeignKey(help_text='The instrument that is this asset.',
                                    on_delete=django.db.models.deletion.PROTECT, related_name='instrument_assets',
                                    to='wbfdm.instrument', verbose_name='Underlying Instrument'),
        ),
        migrations.RunSQL(
            sql="UPDATE wbportfolio_assetposition SET underlying_quote_id = underlying_instrument_id",
            reverse_sql=migrations.RunSQL.noop
        ),
        migrations.AlterField(
            model_name='assetposition',
            name='underlying_quote',
            field=models.ForeignKey(help_text='The quote that is this asset.',
                                    limit_choices_to=models.Q(('children__isnull', True)),
                                    on_delete=django.db.models.deletion.PROTECT, related_name='assets',
                                    to='wbfdm.instrument', verbose_name='Underlying Quote'),
        ),
        migrations.AddIndex(
            model_name='assetposition',
            index=models.Index(fields=['date', 'underlying_instrument', 'portfolio'],
                               name='wbportfolio_date_0a7c0a_idx'),
        ),
        migrations.AddIndex(
            model_name='assetposition',
            index=models.Index(fields=['date', 'underlying_instrument'], name='wbportfolio_date_da82d2_idx'),
        ),
        migrations.AddConstraint(
            model_name='assetposition',
            constraint=models.CheckConstraint(check=models.Q(('date__week_day__in', [1, 7]), _negated=True),
                                              name='wbportfolio_assetposition_weekday_constraint'),
        ),
        migrations.AddConstraint(
            model_name='assetposition',
            constraint=models.UniqueConstraint(fields=('portfolio', 'date', 'underlying_quote', 'portfolio_created'),
                                               name='unique_asset_position', nulls_distinct=False),
        ),
        migrations.RenameField(
            model_name='assetposition',
            old_name='underlying_instrument_price',
            new_name='underlying_quote_price',
        ),
        migrations.AlterField(
            model_name='portfolioportfoliothroughmodel',
            name='type',
            field=models.CharField(choices=[('PRIMARY', 'Primary'), ('MODEL', 'Model'), ('CUSTODIAN', 'Custodian')],
                                   default='PRIMARY', verbose_name='Type'),
        ),
        migrations.AlterField(
            model_name='rebalancer',
            name='parameters',
            field=models.JSONField(blank=True, default=dict, verbose_name='Parameters'),
        ),
        migrations.AddField(
            model_name='portfolio',
            name='is_composition',
            field=models.BooleanField(default=False,
                                      help_text='If true, this portfolio is a composition of other portfolio'),
        ),
        migrations.RunPython(set_composition)
    ]
